from ezdl.models.regseg import RegSeg48
from ezdl.models.backbones import ResNet
from ezdl.models.segnet import SegNet
from ezdl.models.random import Random
from ezdl.models.lawin import Lawin, Laweed, DoubleLawin, DoubleLaweed, SplitLawin, SplitLaweed

MODELS = {
    'segnet': SegNet,
    'regseg48': RegSeg48,
    'random': Random,
    'lawin': Lawin,
    'laweed': Laweed,
    'doublelawin': DoubleLawin,
    "doublelaweed": DoubleLaweed,
    'splitlaweed': SplitLaweed,
    'splitlawin': SplitLawin,
    'resnet': ResNet
}



import torch
from super_gradients.training.utils import get_param, HpmStruct
from super_gradients.training import utils as sg_utils
from torch import Tensor
from torch.nn import functional as F

from ezdl.utils.utilities import filter_none
from .backbones.mit import MiTFusion
from ezdl.models.base import BaseModel
from ezdl.models.heads.lawin import LawinHead
from ezdl.models.heads.laweed import LaweedHead


class BaseLawin(BaseModel):
    """
    Abstract base lawin class with free decoder head lawin based
    """
    def __init__(self, arch_params, lawin_class) -> None:
        num_classes = get_param(arch_params, "num_classes")
        input_channels = get_param(arch_params, "input_channels", 3)
        backbone = get_param(arch_params, "backbone", 'MiT-B0')
        backbone_pretrained = get_param(arch_params, "backbone_pretrained", False)
        pretrained_channels = get_param(arch_params, "main_pretrained", None)
        super().__init__(backbone, input_channels, backbone_pretrained)
        self.decode_head = lawin_class(self.backbone.channels, 256 if 'B0' in backbone else 512, num_classes)
        self.apply(self._init_weights)
        if backbone_pretrained:
            self.main_pretrained = pretrained_channels
            if isinstance(pretrained_channels, str):
                self.main_pretrained = [pretrained_channels] * input_channels
            else:
                self.main_pretrained = pretrained_channels
            self.backbone.init_pretrained_weights(self.main_pretrained)

    def forward(self, x: Tensor) -> Tensor:
        y = self.backbone(x)
        y = self.decode_head(y)   # 4x reduction in image size
        y = F.interpolate(y, size=x.shape[2:], mode='bilinear', align_corners=False)    # to original image shape
        return y


class Lawin(BaseLawin):
    """
    Notes::::: This implementation has larger params and FLOPs than the results reported in the paper.
    Will update the code and weights if the original author releases the full code.
    """
    def __init__(self, arch_params) -> None:
        super().__init__(arch_params, LawinHead)


class Laweed(BaseLawin):
    """
    Notes::::: This implementation has larger params and FLOPs than the results reported in the paper.
    Will update the code and weights if the original author releases the full code.
    """
    def __init__(self, arch_params) -> None:
        super().__init__(arch_params, LaweedHead)


class BaseDoubleLawin(BaseLawin):
    """
    Notes::::: This implementation has larger params and FLOPs than the results reported in the paper.
    Will update the code and weights if the original author releases the full code.
    """
    def __init__(self, arch_params, lawin_class) -> None:
        backbone = get_param(arch_params, "backbone", 'MiT-B0')
        main_channels = get_param(arch_params, "main_channels", None)
        if main_channels is None:
            raise ValueError("Please provide main_channels")
        self.side_channels = arch_params['input_channels'] - main_channels
        self.side_pretrained = get_param(arch_params, "side_pretrained", None)
        self.main_channels = main_channels
        arch_params['input_channels'] = arch_params['main_channels']
        super().__init__(arch_params, lawin_class)
        self.side_backbone = self.eval_backbone(backbone, self.side_channels, pretrained=bool(self.side_pretrained))
        if self.side_pretrained is not None:
            if isinstance(self.side_pretrained, str):
                self.side_pretrained = [self.side_pretrained] * self.side_channels
            self.side_backbone.init_pretrained_weights(self.side_pretrained)
        p_local = get_param(arch_params, "p_local", None)
        p_glob = get_param(arch_params, "p_glob", None)
        fusion_type = get_param(arch_params, "fusion_type", None)
        self.fusion = MiTFusion(self.backbone.channels,
                                **filter_none({"p_local": p_local, "p_glob": p_glob, "fusion_type": fusion_type}))

    def forward(self, x: Tensor) -> Tensor:
        main_channels = x[:, :self.main_channels, ::].contiguous()
        side_channels = x[:, self.main_channels:, ::].contiguous()
        feat_main = self.backbone(main_channels)
        feat_side = self.side_backbone(side_channels)
        feat = self.fusion((feat_main, feat_side))
        y = self.decode_head(feat)   # 4x reduction in image size
        y = F.interpolate(y, size=x.shape[2:], mode='bilinear', align_corners=False)    # to original image shape
        return y


class DoubleLawin(BaseDoubleLawin):
    def __init__(self, arch_params) -> None:
        super().__init__(arch_params, LawinHead)


class DoubleLaweed(BaseDoubleLawin):
    def __init__(self, arch_params) -> None:
        super().__init__(arch_params, LaweedHead)


class BaseSplitLawin(BaseLawin):
    def __init__(self, arch_params, lawin_class) -> None:
        backbone = get_param(arch_params, "backbone", 'MiT-B0')
        main_channels = get_param(arch_params, "main_channels", None)
        if main_channels is None:
            raise ValueError("Please provide main_channels")
        self.side_channels = arch_params['input_channels'] - main_channels
        self.side_pretrained = get_param(arch_params, "side_pretrained", None)
        self.main_channels = main_channels
        arch_params['input_channels'] = arch_params['main_channels']
        super().__init__(arch_params, lawin_class)
        self.side_backbone = self.eval_backbone(backbone, self.side_channels,
                                                n_blocks=1,
                                                pretrained=bool(self.side_pretrained))
        if self.side_pretrained is not None:
            if isinstance(self.side_pretrained, str):
                self.side_pretrained = [self.side_pretrained] * self.side_channels
            self.side_backbone.init_pretrained_weights(self.side_pretrained)
        p_local = get_param(arch_params, "p_local", None)
        p_glob = get_param(arch_params, "p_glob", None)
        fusion_type = get_param(arch_params, "fusion_type", None)
        self.fusion = MiTFusion(self.backbone.channels,
                                **filter_none({"p_local": p_local, "p_glob": p_glob, "fusion_type": fusion_type}))

    def forward(self, x: Tensor) -> Tensor:
        main_channels = x[:, :self.main_channels, ::].contiguous()
        side_channels = x[:, self.main_channels:, ::].contiguous()
        first_feat_side = self.side_backbone(side_channels)
        first_feat_main = self.backbone.partial_forward(main_channels, slice(0, 1))
        first_feat = self.fusion((first_feat_main, first_feat_side))[0]
        feat = (first_feat,) + self.backbone.partial_forward(first_feat, slice(1, 4))
        y = self.decode_head(feat)   # 4x reduction in image size
        y = F.interpolate(y, size=x.shape[2:], mode='bilinear', align_corners=False)    # to original image shape
        return y

    def initialize_param_groups(self, lr: float, training_params: HpmStruct) -> list:
        """

        :return: list of dictionaries containing the key 'named_params' with a list of named params
        """
        def f(x):
            return not (x[0].startswith('backbone') and int(x[0].split('.')[4]) == 0)
        freeze_pretrained = sg_utils.get_param(training_params, 'freeze_pretrained', False)
        if self.backbone_pretrained and freeze_pretrained:
            return [{'named_params': list(filter(f, list(self.named_parameters())))}]
        return [{'named_params': self.named_parameters()}]


class SplitLawin(BaseSplitLawin):
    def __init__(self, arch_params) -> None:
        super().__init__(arch_params, LawinHead)


class SplitLaweed(BaseSplitLawin):
    def __init__(self, arch_params) -> None:
        super().__init__(arch_params, LaweedHead)


if __name__ == '__main__':
    model = Lawin('MiT-B1')
    model.eval()
    x = torch.zeros(1, 3, 512, 512)
    y = model(x)
    print(y.shape)
    from fvcore.nn import flop_count_table, FlopCountAnalysis
    print(flop_count_table(FlopCountAnalysis(model, x)))


import itertools
import os
from typing import Any, Union, Iterable, Mapping

import functools

import numpy as np
import torch
from PIL import Image
from torchvision.datasets.vision import VisionDataset
import torchvision.transforms as transforms
from super_gradients.training.datasets.dataset_interfaces import DatasetInterface
from torchvision.transforms import functional as F

from ezdl.transforms import \
    PairRandomCrop, SegOneHot, ToLong, FixValue, Denormalize, PairRandomFlip, squeeze0, \
    PairFlip, PairFourCrop
from ezdl.data.sequoia_stats import STATS as SEQUOIA_STATS
from ezdl.data.rededge_stats import STATS as REDEDGE_STATS
from sklearn.model_selection import train_test_split

from torch.utils.data.distributed import DistributedSampler
from super_gradients.training import utils as core_utils
from super_gradients.training.datasets.mixup import CollateMixup
from super_gradients.training.exceptions.dataset_exceptions import IllegalDatasetParameterException
from super_gradients.common.abstractions.abstract_logger import get_logger

logger = get_logger(__name__)


class WeedMapDatasetInterface(DatasetInterface):
    STATS = {
        'sequoia': SEQUOIA_STATS,
        'rededge': REDEDGE_STATS
    }

    def __init__(self, dataset_params, name=None):
        super(WeedMapDatasetInterface, self).__init__(dataset_params)
        channels = dataset_params['channels']
        if name == None:
            name = os.path.basename(os.path.normpath(dataset_params['root'])).lower()
        else:
            if name != os.path.basename(os.path.normpath(dataset_params.root)).lower():
                raise ValueError(f'Warning! Dataset name should match the dataset, found: {name}, {dataset_params.root}')
        self.dataset_name = name

        mean, std = self.get_mean_std(dataset_params['train_folders'], channels, name)

        self.lib_dataset_params = {
            'mean': mean,
            'std': std,
            'channels': channels
        }

        input_transform = [
            transforms.Normalize(self.lib_dataset_params['mean'], self.lib_dataset_params['std']),
        ]

        test_transform = [
            transforms.Normalize(self.lib_dataset_params['mean'], self.lib_dataset_params['std']),
        ]

        test_target_transform = [
            transforms.PILToTensor(),
            squeeze0,
            ToLong(),
            FixValue(source=10000, target=1),
            SegOneHot(num_classes=len(WeedMapDataset.CLASS_LABELS.keys()))
        ]

        target_transform = [
            transforms.PILToTensor(),
            squeeze0,
            ToLong(),
            FixValue(source=10000, target=1),
            SegOneHot(num_classes=len(WeedMapDataset.CLASS_LABELS.keys()))
        ]
        period = 1

        crop_size = core_utils.get_param(self.dataset_params, 'crop_size', default_val='same')
        if crop_size != 'same':
            crop = PairRandomCrop(crop_size)
            test_crop = PairFourCrop(crop_size, periodicity=period)

            input_transform.append(crop)
            target_transform.append(crop)
            test_transform.append(test_crop)
            test_target_transform.append(test_crop)
            period *= 4

        if dataset_params['hor_flip']:
            flip_hor = PairRandomFlip(orientation="horizontal")

            input_transform.append(flip_hor)
            target_transform.append(flip_hor)

        if dataset_params['ver_flip']:
            flip_ver = PairRandomFlip(orientation="vertical")

            input_transform.append(flip_ver)
            target_transform.append(flip_ver)

        if core_utils.get_param(self.dataset_params, 'size', default_val='same') != 'same':
            resize = transforms.Resize(size=core_utils.get_param(self.dataset_params, 'size', default_val='same'))

            input_transform.append(resize)
            target_transform.append(resize)
            test_transform.append(resize)
            test_target_transform.append(resize)

        target_transform = transforms.Compose(target_transform)
        input_transform = transforms.Compose(input_transform)
        test_transform = transforms.Compose(test_transform)
        test_target_transform = transforms.Compose(test_target_transform)

        # Divide train, val and test
        self.train_folders = dataset_params['train_folders']
        self.test_folders = dataset_params['test_folders']

        train_index = WeedMapDataset.build_index(self.dataset_params.root, self.train_folders, channels)
        test_index = WeedMapDataset.build_index(self.dataset_params.root, self.test_folders, channels)
        train_index, val_index = train_test_split(train_index, test_size=0.2)

        self.trainset = WeedMapDataset(root=self.dataset_params.root, channels=channels,
                                       batch_size=self.dataset_params.batch_size, index=train_index,
                                       transform=input_transform, target_transform=target_transform,
                                       return_path=dataset_params['return_path'])

        self.valset = WeedMapDataset(root=self.dataset_params.root, channels=channels,
                                     batch_size=self.dataset_params.val_batch_size, index=val_index,
                                     transform=input_transform, target_transform=target_transform,
                                     return_path=dataset_params['return_path'])

        self.testset = WeedMapDataset(root=self.dataset_params.root, channels=channels,
                                      batch_size=self.dataset_params.test_batch_size, index=test_index,
                                      transform=test_transform, target_transform=test_target_transform,
                                      return_path=dataset_params['return_path'], period=period)

    def undo_preprocess(self, x):
        return (Denormalize(self.lib_dataset_params['mean'], self.lib_dataset_params['std'])(x) * 255).type(torch.uint8)

    @classmethod
    def get_mean_std(cls, train_folders, channels, dataset_name):
        stats = cls.STATS[dataset_name]
        if channels == 'CIR':
            channels = ['NIR', 'G', 'R']
        if len(train_folders) == 1:
            f = train_folders[0]
            return list(zip(*[(stats[f][c]['mean'], stats[f][c]['std']) for c in channels]))
        else:
            sums = {
                **{c + '_sum': sum([stats[f][c]['sum'] for f in train_folders]) for c in channels},
                **{c + '_sum_sq': sum([stats[f][c]['sum_sq'] for f in train_folders]) for c in channels}
            }
            count = sum([stats[f]['count'] for f in train_folders])
            means = [sums[c + '_sum'] / count for c in channels]
            stds = [np.sqrt((sums[c + '_sum_sq'] / count) - (means[i] ** 2))
                    for i, c in enumerate(channels)]
            return means, stds

    def build_data_loaders(self, batch_size_factor=1, num_workers=8, train_batch_size=None, val_batch_size=None,
                           test_batch_size=None, distributed_sampler: bool = False):
        """

        define train, val (and optionally test) loaders. The method deals separately with distributed training and standard
        (non distributed, or parallel training). In the case of distributed training we need to rely on distributed
        samplers.
        :param batch_size_factor: int - factor to multiply the batch size (usually for multi gpu)
        :param num_workers: int - number of workers (parallel processes) for dataloaders
        :param train_batch_size: int - batch size for train loader, if None will be taken from dataset_params
        :param val_batch_size: int - batch size for val loader, if None will be taken from dataset_params
        :param distributed_sampler: boolean flag for distributed training mode
        :return: train_loader, val_loader, classes: list of classes
        """
        # CHANGE THE BATCH SIZE ACCORDING TO THE NUMBER OF DEVICES - ONLY IN NON-DISTRIBUED TRAINING MODE
        # IN DISTRIBUTED MODE WE NEED DISTRIBUTED SAMPLERS
        # NO SHUFFLE IN DISTRIBUTED TRAINING
        if distributed_sampler:
            self.batch_size_factor = 1
            train_sampler = DistributedSampler(self.trainset)
            val_sampler = DistributedSampler(self.valset)
            test_sampler = DistributedSampler(self.testset) if self.testset is not None else None
            train_shuffle = False
        else:
            self.batch_size_factor = batch_size_factor
            train_sampler = None
            val_sampler = None
            test_sampler = None
            train_shuffle = True

        if train_batch_size is None:
            train_batch_size = self.dataset_params.batch_size * self.batch_size_factor
        if val_batch_size is None:
            val_batch_size = self.dataset_params.val_batch_size * self.batch_size_factor
        if test_batch_size is None:
            test_batch_size = self.dataset_params.test_batch_size * self.batch_size_factor

        train_loader_drop_last = core_utils.get_param(self.dataset_params, 'train_loader_drop_last', default_val=False)

        cutmix = core_utils.get_param(self.dataset_params, 'cutmix', False)
        cutmix_params = core_utils.get_param(self.dataset_params, 'cutmix_params')

        # WRAPPING collate_fn
        train_collate_fn = core_utils.get_param(self.trainset, 'collate_fn')
        val_collate_fn = core_utils.get_param(self.valset, 'collate_fn')
        test_collate_fn = core_utils.get_param(self.testset, 'collate_fn')

        if cutmix and train_collate_fn is not None:
            raise IllegalDatasetParameterException("cutmix and collate function cannot be used together")

        if cutmix:
            # FIXME - cutmix should be available only in classification dataset. once we make sure all classification
            # datasets inherit from the same super class, we should move cutmix code to that class
            logger.warning("Cutmix/mixup was enabled. This feature is currently supported only "
                           "for classification datasets.")
            train_collate_fn = CollateMixup(**cutmix_params)

        # FIXME - UNDERSTAND IF THE num_replicas VARIBALE IS NEEDED
        # train_sampler = DistributedSampler(self.trainset,
        #                                    num_replicas=distributed_gpus_num) if distributed_sampler else None
        # val_sampler = DistributedSampler(self.valset,
        #                                   num_replicas=distributed_gpus_num) if distributed_sampler else None
        pw = num_workers > 0
        self.train_loader = torch.utils.data.DataLoader(self.trainset,
                                                        batch_size=train_batch_size,
                                                        shuffle=train_shuffle,
                                                        num_workers=num_workers,
                                                        pin_memory=True,
                                                        sampler=train_sampler,
                                                        collate_fn=train_collate_fn,
                                                        drop_last=train_loader_drop_last,
                                                        persistent_workers=pw)

        self.val_loader = torch.utils.data.DataLoader(self.valset,
                                                      batch_size=val_batch_size,
                                                      shuffle=False,
                                                      num_workers=num_workers,
                                                      pin_memory=True,
                                                      sampler=val_sampler,
                                                      collate_fn=val_collate_fn,
                                                      persistent_workers=pw)

        if self.testset is not None:
            self.test_loader = torch.utils.data.DataLoader(self.testset,
                                                           batch_size=test_batch_size,
                                                           shuffle=False,
                                                           num_workers=num_workers,
                                                           pin_memory=True,
                                                           sampler=test_sampler,
                                                           collate_fn=test_collate_fn,
                                                           persistent_workers=pw)

        self.classes = self.trainset.classes

    def get_run_loader(self, root=None, folders=None, batch_size=None):
        batch_size = batch_size if batch_size is not None else self.dataset_params.test_batch_size
        folders = folders if folders is not None else self.test_folders
        root = root if root is not None else self.dataset_params.root

        index = WeedMapDataset.build_index(root, folders, self.dataset_params.channels)
        input_transform = [
            transforms.Normalize(self.lib_dataset_params['mean'], self.lib_dataset_params['std']),
        ]

        target_transform = [
            transforms.PILToTensor(),
            squeeze0,
            ToLong(),
            FixValue(source=10000, target=1),
            SegOneHot(num_classes=len(WeedMapDataset.CLASS_LABELS.keys()))
        ]

        if core_utils.get_param(self.dataset_params, 'size', default_val='same') != 'same':
            resize = transforms.Resize(size=core_utils.get_param(self.dataset_params, 'size', default_val='same'))
            input_transform.append(resize)
            target_transform.append(resize)

        target_transform = transforms.Compose(target_transform)
        input_transform = transforms.Compose(input_transform)

        runset = WeedMapDataset(root=self.dataset_params.root, channels=self.dataset_params.channels,
                                batch_size=self.dataset_params.test_batch_size, index=index,
                                transform=input_transform, target_transform=target_transform)

        loader = torch.utils.data.DataLoader(runset,
                                             batch_size=batch_size,
                                             num_workers=self.dataset_params.num_workers,
                                             pin_memory=True,
                                             persistent_workers=False)
        return loader


class WeedMapDataset(VisionDataset):
    CLASS_LABELS = {0: "background", 1: "crop", 2: 'weed'}
    classes = ['background', 'crop', 'weed']

    def __init__(self,
                 root: str,
                 transform: callable,
                 target_transform: callable,
                 channels: Union[str, Iterable] = 'CIR',
                 index: Iterable = None,
                 batch_size: int = 4,
                 return_mask: bool = False,
                 return_path: bool = False,
                 period: int = None,
                 ):
        """
        Initialize a SequoiaDataset object.

        :param batch_size: The batch size.
        :param root: The root directory.
        :param transform: The transform to apply to the data.
        :param target_transform: The transform to apply to the target.
        :param channels: The channels to use.
        :param index: The index to get the images.
        :param return_mask: Whether to return the mask.
        :param return_path: Whether to return the path.
        :param period: Used when augmentation returns different augmentations for each image.
        """
        self.batch_size = batch_size
        super().__init__(root=root)

        if channels == 'CIR' or len(channels) == 1:
            channels = channels[0] if len(channels) == 1 else channels
            self.get_img = self.get_complete_image
        else:
            self.get_img = self.get_channels

        if index:
            self.index = index
        else:
            self.index = self.build_index(root, channels=channels)

        if period is not None:
            self.index = list(itertools.chain.from_iterable(itertools.repeat(x, period) for x in self.index))

        self.len = len(self.index)

        self.channels = channels
        self.n_files = {}
        self.path = root
        self.transform = transform
        self.target_transform = target_transform
        self.return_mask = return_mask
        self.return_name = return_path

    @classmethod
    def build_index(cls,
                    root: str,
                    macro_folders: Iterable = None,
                    channels: Union[Iterable, str] = 'CIR') -> list:

        if macro_folders is None:
            macro_folders = os.listdir(root)

        if channels == 'CIR':
            counter_channel = channels
        else:
            counter_channel = channels[0]  # Channel used to count images

        n_files = dict()
        for folder in macro_folders:
            n_files[folder] = (os.listdir(os.path.join(root, folder, 'tile', counter_channel)))
        index = dict()
        i = 0
        for folder, files in n_files.items():
            for file in files:
                index[i] = folder, file
                i += 1
        return list(index.values())

    def get_complete_image(self, folder, file) -> Any:
        img = Image.open(
            os.path.join(self.path, folder, 'tile', self.channels, file)
        )
        return F.to_tensor(img)

    def get_channels(self, folder, file) -> Any:
        return torch.stack([
            F.to_tensor(Image.open(
                os.path.join(self.path, folder, 'tile', c, file)
            )).squeeze(0) for c in self.channels
        ]
        )

    def __getitem__(self, index: int) -> Any:
        """
        Get the i-th image and related target

        :param idx: The index of the image.
        :return: A pair consisting of the image and the target
        """
        folder, file = self.index[index]
        img = self.get_img(folder, file)
        gt = Image.open(
            os.path.join(self.path, folder, 'groundtruth',
                         folder + '_' + file.split('.')[0] + '_GroundTruth_iMap.png'
                         )
        )
        img = self.transform(img)
        gt = self.target_transform(gt)

        if self.return_mask:
            mask = Image.open(
                os.path.join(self.path, folder, 'mask', file)
            )
            gt = (gt, mask)
        if self.return_name:
            return img, gt, {'input_name': folder + '_' + file}
        return img, gt

    def __len__(self) -> int:
        """
        Get the number of batches.

        :return: The number of batches.
        """
        return self.len



import torch
from torch.nn import functional as F

from .lawin import LawinAttn, ConvModule, PatchEmbed, LawinHead

LOW_LEVEL_PATCH_SIZE = 8
LOW_LEVEL_RATIO = 16


class LaweedHead(LawinHead):
    def __init__(self, in_channels: list, embed_dim=512, num_classes=19) -> None:
        super().__init__(in_channels, embed_dim, num_classes)

        setattr(self, f"lawin_low_{LOW_LEVEL_RATIO}",
                LawinAttn(embed_dim // 8, 16, patch_size=LOW_LEVEL_PATCH_SIZE))
        setattr(self, f"ds_low_{LOW_LEVEL_RATIO}", PatchEmbed(LOW_LEVEL_RATIO, embed_dim // 8, embed_dim // 8))
        self.low_cat = ConvModule(in_channels[0]*2, in_channels[0])

    def forward(self, features):
        B, _, H, W = features[1].shape
        outs = [self.linear_c2(features[1]).permute(0, 2, 1).reshape(B, -1, *features[1].shape[-2:])]

        for i, feature in enumerate(features[2:]):
            cf = eval(f"self.linear_c{i+3}")(feature).permute(0, 2, 1).reshape(B, -1, *feature.shape[-2:])
            outs.append(F.interpolate(cf, size=(H, W), mode='bilinear', align_corners=False))

        feat = self.linear_fuse(torch.cat(outs[::-1], dim=1))
        B, _, H, W = feat.shape

        ## Lawin attention spatial pyramid pooling
        feat_short = self.short_path(feat)
        feat_pool = F.interpolate(self.image_pool(feat), size=(H, W), mode='bilinear', align_corners=False)
        feat_lawin = self.get_lawin_att_feats(feat, 8, self.ratios)
        output = self.cat(torch.cat([feat_short, feat_pool, *feat_lawin], dim=1))

        ## Low-level feature enhancement
        low_attn = self.get_lawin_att_feats(features[0], LOW_LEVEL_PATCH_SIZE, [LOW_LEVEL_RATIO], step="low_")[0]
        low_cat = self.low_cat(torch.cat([low_attn, features[0]], dim=1))
        c1 = self.linear_c1(low_cat).permute(0, 2, 1).reshape(B, -1, *features[0].shape[-2:])
        output = F.interpolate(output, size=features[0].shape[-2:], mode='bilinear', align_corners=False)
        fused = self.low_level_fuse(torch.cat([output, c1], dim=1))

        seg = self.linear_pred(self.dropout(fused))
        return seg



import torch 
from torch import nn, Tensor
from torch.nn import functional as F


class BasicBlock(nn.Module):
    """2 Layer No Expansion Block
    """
    expansion: int = 1
    def __init__(self, c1, c2, s=1, downsample= None) -> None:
        super().__init__()
        self.conv1 = nn.Conv2d(c1, c2, 3, s, 1, bias=False)
        self.bn1 = nn.BatchNorm2d(c2)
        self.conv2 = nn.Conv2d(c2, c2, 3, 1, 1, bias=False)
        self.bn2 = nn.BatchNorm2d(c2)
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        identity = x
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        if self.downsample is not None: identity = self.downsample(x)
        out += identity
        return F.relu(out)


class Bottleneck(nn.Module):
    """3 Layer 4x Expansion Block
    """
    expansion: int = 4
    def __init__(self, c1, c2, s=1, downsample=None) -> None:
        super().__init__()
        self.conv1 = nn.Conv2d(c1, c2, 1, 1, 0, bias=False)
        self.bn1 = nn.BatchNorm2d(c2)
        self.conv2 = nn.Conv2d(c2, c2, 3, s, 1, bias=False)
        self.bn2 = nn.BatchNorm2d(c2)
        self.conv3 = nn.Conv2d(c2, c2 * self.expansion, 1, 1, 0, bias=False)
        self.bn3 = nn.BatchNorm2d(c2 * self.expansion)
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        identity = x
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        if self.downsample is not None: identity = self.downsample(x)
        out += identity
        return F.relu(out)


resnet_settings = {
    '18': [BasicBlock, [2, 2, 2, 2], [64, 128, 256, 512]],
    '34': [BasicBlock, [3, 4, 6, 3], [64, 128, 256, 512]],
    '50': [Bottleneck, [3, 4, 6, 3], [256, 512, 1024, 2048]],
    '101': [Bottleneck, [3, 4, 23, 3], [256, 512, 1024, 2048]],
    '152': [Bottleneck, [3, 8, 36, 3], [256, 512, 1024, 2048]]
}


class ResNet(nn.Module):
    def __init__(self, args) -> None:
        model_name = args.get('model_name') or '50'
        super().__init__()
        assert model_name in resnet_settings.keys(), f"ResNet model name should be in {list(resnet_settings.keys())}"
        block, depths, channels = resnet_settings[model_name]

        self.inplanes = 64
        self.channels = channels
        self.conv1 = nn.Conv2d(3, self.inplanes, 7, 2, 3, bias=False)
        self.bn1 = nn.BatchNorm2d(self.inplanes)
        self.maxpool = nn.MaxPool2d(3, 2, 1)

        self.layer1 = self._make_layer(block, 64, depths[0], s=1)
        self.layer2 = self._make_layer(block, 128, depths[1], s=2)
        self.layer3 = self._make_layer(block, 256, depths[2], s=2)
        self.layer4 = self._make_layer(block, 512, depths[3], s=2)


    def _make_layer(self, block, planes, depth, s=1) -> nn.Sequential:
        downsample = None
        if s != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion, 1, s, bias=False),
                nn.BatchNorm2d(planes * block.expansion)
            )
        layers = nn.Sequential(
            block(self.inplanes, planes, s, downsample),
            *[block(planes * block.expansion, planes) for _ in range(1, depth)]
        )
        self.inplanes = planes * block.expansion
        return layers


    def forward(self, x: Tensor) -> Tensor:
        x = self.maxpool(F.relu(self.bn1(self.conv1(x))))   # [1, 64, H/4, W/4]
        x1 = self.layer1(x)  # [1, 64/256, H/4, W/4]   
        x2 = self.layer2(x1)  # [1, 128/512, H/8, W/8]
        x3 = self.layer3(x2)  # [1, 256/1024, H/16, W/16]
        x4 = self.layer4(x3)  # [1, 512/2048, H/32, W/32]
        return x1, x2, x3, x4


if __name__ == '__main__':
    model = ResNet('18')
    # model.load_state_dict(torch.load('C:\\Users\\sithu\\Documents\\weights\\backbones\\resnet\\resnet18_a1.pth', map_location='cpu'), strict=False)
    x = torch.zeros(1, 3, 224, 224)
    outs = model(x)
    for y in outs:
        print(y.shape) 


import copy
import gc
import os
from typing import Mapping
from easydict import EasyDict

from ezdl.experiment.resume import ExpLog
from ezdl.experiment.run import Run
from ezdl.experiment.resume import get_interrupted_run, retrieve_run_to_resume
from ezdl.utils.grid import make_grid, linearize
from ezdl.utils.utilities import nested_dict_update, update_collection
from super_gradients.common.abstractions.abstract_logger import get_logger

logger = get_logger(__name__)


class GridSummary:
    def __init__(self,
                 total_runs,
                 total_runs_excl_grid,
                 total_runs_to_run,
                 total_runs_excl,
                 ):
        self.total_runs = total_runs
        self.total_runs_excl_grid = total_runs_excl_grid
        self.total_runs_to_run = total_runs_to_run
        self.total_runs_excl = total_runs_excl

    def update(self, d):
        self.total_runs = d.get("total_runs") or self.total_runs
        self.total_runs_excl_grid = d.get("total_runs_excl_grid") or self.total_runs_excl_grid
        self.total_runs_to_run = d.get("total_runs_to_run") or self.total_runs_to_run
        self.total_runs_excl = d.get("total_runs_excl") or self.total_runs_to_run


class ExpSettings(EasyDict):
    def __init__(self, *args, **kwargs):
        self.start_from_grid = 0
        self.start_from_run = 0
        self.resume = False
        self.resume_last = False
        self.tracking_dir = ""
        self.excluded_files = ""
        super().__init__(*args, **kwargs)


class Experimenter:
    def __init__(self):
        self.gs = None
        self.exp_settings = ExpSettings()
        self.grids = None

    def calculate_runs(self, settings):
        base_grid = settings['parameters']
        other_grids = settings['other_grids']
        self.exp_settings = ExpSettings(settings['experiment'])

        complete_grids = [base_grid]
        if other_grids:
            complete_grids += \
                [nested_dict_update(copy.deepcopy(base_grid), other_run) for other_run in other_grids]
        logger.info(f'There are {len(complete_grids)} grids')

        self.grids, dot_elements = zip(*[make_grid(grid, return_cartesian_elements=True) for grid in complete_grids])
        dot_elements = list(dot_elements)
        dot_elements[1:] = [list(dict(linearize(others) + dot).items()) for others, dot in
                            zip(other_grids, dot_elements[1:])]

        # Modify starting grid and run to manage the resume
        self.manage_resume()

        for i, grid in enumerate(self.grids):
            info = f'Found {len(grid)} runs from grid {i}'
            if i < self.exp_settings.start_from_grid:
                info += f', skipping grid {i} with {len(grid)} runs'
            logger.info(info)
        self.generate_grid_summary()
        return self.gs, self.grids, dot_elements

    def generate_grid_summary(self):
        total_runs = sum(len(grid) for grid in self.grids)
        total_runs_excl_grid = total_runs - sum([len(grid) for grid in self.grids[self.exp_settings.start_from_grid:]])
        total_runs_excl = total_runs_excl_grid + self.exp_settings.start_from_run
        total_runs_to_run = total_runs - total_runs_excl
        self.gs = GridSummary(
            total_runs=total_runs,
            total_runs_excl_grid=total_runs_excl_grid,
            total_runs_to_run=total_runs_to_run,
            total_runs_excl=total_runs_excl
        )

    def execute_runs(self, callback=None):
        track_dir = self.exp_settings['tracking_dir']
        exp_log = ExpLog(track_dir)
        exp_log.start()
        starting_run = self.exp_settings.start_from_grid
        if self.exp_settings.resume_last:
            logger.info("+ another run to finish!")
            grid_len = len(self.grids[self.exp_settings.start_from_grid])
            sg = self.exp_settings.start_from_grid
            sr = self.exp_settings.start_from_run - 1
            try:
                exp_log.insert_run(sg, sr)
                run = get_interrupted_run(self.exp_settings)
                if callback:
                    yield callback(self.exp_settings.start_from_grid, self.exp_settings.start_from_run - 1,
                                   len(self.grids), grid_len,
                                   status="started", run_params=run.params)
                logger.info(f'Running grid {sg} out of {len(self.grids) - 1}')
                logger.info(f'Running run {sr - 1} out of {grid_len} '
                            f'({sum([len(self.grids[k]) for k in range(sg)]) + sr} / {self.gs.total_runs - 1})')
                run.launch()
                exp_log.finish_run()
            except Exception as e:
                logger.error(f'Experiment {sg} failed with error {e}')
                exp_log.crash_run()
                if not self.exp_settings.continue_with_errors:
                    raise e
                if callback:
                    yield callback(sg, sr, len(self.grids), grid_len, status="crashed", run_params={}, exception=e)
        for i in range(self.exp_settings.start_from_grid, len(self.grids)):
            grid = self.grids[i]
            if i != self.exp_settings.start_from_grid:
                starting_run = 0
            for j in range(starting_run, len(grid)):
                params = grid[j]
                try:
                    exp_log.insert_run(i, j)
                    if callback:
                        yield callback(i, j, len(self.grids), len(grid), status="started", run_params=params)
                    logger.info(f'Running grid {i} out of {len(self.grids) - 1}')
                    logger.info(f'Running run {j} out of {len(grid) - 1} '
                                f'({sum([len(self.grids[k]) for k in range(i)]) + j} / {self.gs.total_runs - 1})')
                    run = Run()
                    run.init({'experiment': {**self.exp_settings}, **params})
                    run.launch()
                    exp_log.finish_run()
                    gc.collect()
                    if callback:
                        yield callback(i, j, len(self.grids), len(grid), status="finished", run_params={})
                except Exception as e:
                    logger.error(f'Experiment {i} failed with error {e}')
                    exp_log.crash_run()
                    if not self.exp_settings.continue_with_errors:
                        raise e
                    if callback:
                        yield callback(i, j, len(self.grids), len(grid), status="crashed", run_params={}, exception=e)
        exp_log.close()

    def manage_resume(self):
        if self.exp_settings.resume:
            self.exp_settings.start_from_grid, \
                self.exp_settings.start_from_run, \
                self.exp_settings.resume_last = retrieve_run_to_resume(self.exp_settings, self.grids)
        else:
            self.exp_settings.resume_last = False

    def update_settings(self, d):
        self.exp_settings = update_collection(self.exp_settings, d)
        if self.gs is None:
            return
        self.gs.update(self.exp_settings)
        if "resume" in d:
            self.manage_resume()
            self.generate_grid_summary()


def experiment(settings: Mapping, param_path: str = "local variable"):
    exp = Experimenter()
    exp.calculate_runs(settings)
    exp_settings = exp.exp_settings

    if exp_settings.excluded_files:
        os.environ['WANDB_IGNORE_GLOBS'] = exp_settings.excluded_files

    logger.info(f'Loaded parameters from {param_path}')

    experimenter = Experimenter()
    grid_summary, grids, cartesian_elements = experimenter.calculate_runs(settings)

    logger.info(f'Total runs found:              {grid_summary.total_runs}')
    logger.info(f'Total runs excluded by grids:  {grid_summary.total_runs_excl_grid}')
    logger.info(f'Total runs excluded:           {grid_summary.total_runs_excl}')
    logger.info(f'Total runs to run:             {grid_summary.total_runs_to_run}')
    experimenter.execute_runs()



import torch
from torch import nn, Tensor
from torch.nn import functional as F
from torchvision.ops import DeformConv2d
from ezdl.models.layers import ConvModule


class DCNv2(nn.Module):
    def __init__(self, c1, c2, k, s, p, g=1):
        super().__init__()
        self.dcn = DeformConv2d(c1, c2, k, s, p, groups=g)
        self.offset_mask = nn.Conv2d(c2,  g* 3 * k * k, k, s, p)
        self._init_offset()

    def _init_offset(self):
        self.offset_mask.weight.data.zero_()
        self.offset_mask.bias.data.zero_()

    def forward(self, x, offset):
        out = self.offset_mask(offset)
        o1, o2, mask = torch.chunk(out, 3, dim=1)
        offset = torch.cat([o1, o2], dim=1)
        mask = mask.sigmoid()
        return self.dcn(x, offset, mask)


class FSM(nn.Module):
    def __init__(self, c1, c2):
        super().__init__()
        self.conv_atten = nn.Conv2d(c1, c1, 1, bias=False)
        self.conv = nn.Conv2d(c1, c2, 1, bias=False)

    def forward(self, x: Tensor) -> Tensor:
        atten = self.conv_atten(F.avg_pool2d(x, x.shape[2:])).sigmoid()
        feat = torch.mul(x, atten)
        x = x + feat
        return self.conv(x)


class FAM(nn.Module):
    def __init__(self, c1, c2):
        super().__init__()
        self.lateral_conv = FSM(c1, c2)
        self.offset = nn.Conv2d(c2*2, c2, 1, bias=False)
        self.dcpack_l2 = DCNv2(c2, c2, 3, 1, 1, 8)
    
    def forward(self, feat_l, feat_s):
        feat_up = feat_s
        if feat_l.shape[2:] != feat_s.shape[2:]:
            feat_up = F.interpolate(feat_s, size=feat_l.shape[2:], mode='bilinear', align_corners=False)
        
        feat_arm = self.lateral_conv(feat_l)
        offset = self.offset(torch.cat([feat_arm, feat_up*2], dim=1))

        feat_align = F.relu(self.dcpack_l2(feat_up, offset))
        return feat_align + feat_arm


class FaPNHead(nn.Module):
    def __init__(self, in_channels, channel=128, num_classes=19):
        super().__init__()
        in_channels = in_channels[::-1]
        self.align_modules = nn.ModuleList([ConvModule(in_channels[0], channel, 1)])
        self.output_convs = nn.ModuleList([])

        for ch in in_channels[1:]:
            self.align_modules.append(FAM(ch, channel))
            self.output_convs.append(ConvModule(channel, channel, 3, 1, 1))

        self.conv_seg = nn.Conv2d(channel, num_classes, 1)
        self.dropout = nn.Dropout2d(0.1)

    def forward(self, features) -> Tensor:
        features = features[::-1]
        out = self.align_modules[0](features[0])
        
        for feat, align_module, output_conv in zip(features[1:], self.align_modules[1:], self.output_convs):
            out = align_module(feat, out)
            out = output_conv(out)
        out = self.conv_seg(self.dropout(out))
        return out


if __name__ == '__main__':
    from semseg.models.backbones import ResNet
    backbone = ResNet('50')
    head = FaPNHead([256, 512, 1024, 2048], 128, 19)
    x = torch.randn(2, 3, 224, 224)
    features = backbone(x)
    out = head(features)
    out = F.interpolate(out, size=x.shape[-2:], mode='bilinear', align_corners=False)
    print(out.shape)


import multiprocessing

import torch
import json

from torch.utils.data import DataLoader
from tqdm import tqdm
from torchvision.transforms import transforms

from ezdl.data.sequoia import WeedMapDataset

WORKERS = multiprocessing.cpu_count()
BATCH_SIZE = 256
WIDTH = 360
HEIGHT = 480


def calculate(root, folders, channels):
    """
    Calculate the mean and the standard deviation of a dataset
    """

    augs = transforms.Compose([
                      transforms.ToTensor()])

    index = WeedMapDataset.build_index(
        root,
        macro_folders=folders,
        channels=channels,
    )

    sq = WeedMapDataset(root,
                        transform=lambda x: x,
                        target_transform=augs,
                        index=index,
                        channels=channels
                        )
    count = len(sq) * WIDTH * HEIGHT


    # data loader
    image_loader = DataLoader(sq,
                              batch_size=BATCH_SIZE,
                              shuffle=False,
                              num_workers=0,
                              pin_memory=True)

    # placeholders
    psum = torch.zeros(len(channels))
    psum_sq = torch.zeros(len(channels))

    # loop through images
    for input, _ in tqdm(image_loader):
        psum += input.sum(axis=[0, 2, 3])
        psum_sq += (input ** 2).sum(axis=[0, 2, 3])

    total_mean = psum / count
    total_var = (psum_sq / count) - (total_mean ** 2)
    total_std = torch.sqrt(total_var)

    d = {}
    for i in range(len(channels)):
        d[channels[i]] = {
            'mean': total_mean[i].item(),
            'std': total_std[i].item(),
            'sum': psum[i].item(),
            'sum_sq': psum_sq[i].item(),
        }
    d['count'] = count
    return d


if __name__ == '__main__':
    # SEQUOIA
    # root = "./dataset/processed/Sequoia"
    # folders = ['005', '006', '007']
    # channels = ['R', 'G', 'NDVI', 'NIR', 'RE']

    # REDEDGE
    root = "./dataset/processed/RedEdge"
    folders = ['000', '001', '002', '004']
    channels = ['R', 'G', 'B', 'NDVI', 'NIR', 'RE']
    d = {}
    for folder in folders:
        d[folder] = calculate(root, [folder], channels)
    print(json.dumps(d, indent=4))


from typing import Tuple, List, Any, Tuple

import torch
from torch import nn, Tensor
from torch.nn import functional as F
from ezdl.models.layers import DropPath, ConvModule, MultiDropPath
from transformers import SegformerModel, SegformerConfig


CHANNEL_PRETRAIN = {'R': 0, 'G': 1, 'B': 2}


class Attention(nn.Module):
    def __init__(self, dim, head, sr_ratio):
        super().__init__()
        self.head = head
        self.sr_ratio = sr_ratio
        self.scale = (dim // head) ** -0.5
        self.q = nn.Linear(dim, dim)
        self.kv = nn.Linear(dim, dim * 2)
        self.proj = nn.Linear(dim, dim)

        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, sr_ratio, sr_ratio)
            self.norm = nn.LayerNorm(dim)

    def forward(self, x: Tensor, H, W) -> Tensor:
        B, N, C = x.shape
        q = self.q(x).reshape(B, N, self.head, C // self.head).permute(0, 2, 1, 3)

        if self.sr_ratio > 1:
            x = x.permute(0, 2, 1).reshape(B, C, H, W)
            x = self.sr(x).reshape(B, C, -1).permute(0, 2, 1)
            x = self.norm(x)

        k, v = self.kv(x).reshape(B, -1, 2, self.head, C // self.head).permute(2, 0, 3, 1, 4)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        return x


class DWConv(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dwconv = nn.Conv2d(dim, dim, 3, 1, 1, groups=dim)

    def forward(self, x: Tensor, H, W) -> Tensor:
        B, _, C = x.shape
        x = x.transpose(1, 2).view(B, C, H, W)
        x = self.dwconv(x)
        return x.flatten(2).transpose(1, 2)


class MLP(nn.Module):
    def __init__(self, c1, c2):
        super().__init__()
        self.fc1 = nn.Linear(c1, c2)
        self.dwconv = DWConv(c2)
        self.fc2 = nn.Linear(c2, c1)

    def forward(self, x: Tensor, H, W) -> Tensor:
        return self.fc2(F.gelu(self.dwconv(self.fc1(x), H, W)))


class PatchEmbed(nn.Module):
    def __init__(self, c1=3, c2=32, patch_size=7, stride=4):
        super().__init__()
        self.proj = nn.Conv2d(c1, c2, patch_size, stride, patch_size // 2)  # padding=(ps[0]//2, ps[1]//2)
        self.norm = nn.LayerNorm(c2)

    def forward(self, x: Tensor) -> Tuple[Any, Any, Any]:
        x = self.proj(x)
        _, _, H, W = x.shape
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)
        return x, H, W


class Block(nn.Module):
    def __init__(self, dim, head, sr_ratio=1, dpr=0.):
        super().__init__()
        self.norm1 = nn.LayerNorm(dim)
        self.attn = Attention(dim, head, sr_ratio)
        self.drop_path = DropPath(dpr) if dpr > 0. else nn.Identity()
        self.norm2 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, int(dim * 4))

    def forward(self, x: Tensor, H, W) -> Tensor:
        x = x + self.drop_path(self.attn(self.norm1(x), H, W))
        x = x + self.drop_path(self.mlp(self.norm2(x), H, W))
        return x


mit_settings = {
    'B0': [[32, 64, 160, 256], [2, 2, 2, 2]],  # [embed_dims, depths]
    'B1': [[64, 128, 320, 512], [2, 2, 2, 2]],
    'B2': [[64, 128, 320, 512], [3, 4, 6, 3]],
    'B3': [[64, 128, 320, 512], [3, 4, 18, 3]],
    'B4': [[64, 128, 320, 512], [3, 8, 27, 3]],
    'B5': [[64, 128, 320, 512], [3, 6, 40, 3]]
}


class MiT(nn.Module):
    pretrained_url = ("nvidia/segformer-", "-finetuned-ade-512-512")

    def __init__(self, model_name: str = 'B0', input_channels=3, n_blocks=4, pretrained=False):
        super().__init__()
        if not pretrained:
            assert model_name in mit_settings.keys(), f"MiT model name should be in {list(mit_settings.keys())}"
            embed_dims, depths = mit_settings[model_name]
            drop_path_rate = 0.1
            self.channels = embed_dims

            # patch_embed
            patch_sizes = [7, 3, 3, 3]
            paddings = [4, 2, 2, 2]
            for i in range(n_blocks):
                c_in = input_channels if i == 0 else embed_dims[i-1]
                setattr(self, f"patch_embed{i+1}", PatchEmbed(c_in, embed_dims[i], patch_sizes[i], paddings[i]))

            dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]

            heads = [1, 2, 5, 8]
            ratios = [8, 4, 2, 1]
            cur = 0
            for i in range(n_blocks):
                setattr(self, f"block{i+1}",
                        nn.ModuleList([Block(embed_dims[i], heads[i], ratios[i], dpr[cur + k]) for k in range(depths[i])]))
                setattr(self, f"norm{i+1}", nn.LayerNorm(embed_dims[i]))
                cur += depths[i]
            self.forward = self.base_forward
            self.partial_forward = self.base_partial_forward
            self.n_blocks = n_blocks
        else:
            url = self.pretrained_url[0] + model_name.lower() + self.pretrained_url[1]
            self.url = url
            config = SegformerConfig().from_pretrained(url)
            config.num_encoder_blocks = n_blocks
            config.num_channels = input_channels
            self.config = config
            self.channels = config.hidden_sizes
            self.encoder = SegformerModel(config)
            self.forward = self.hug_forward
            self.partial_forward = self.hug_partial_forward
            self.n_blocks = n_blocks

    def init_pretrained_weights(self, channel_to_load=None):
        if channel_to_load is None:
            channel_to_load = slice(self.config.num_channels)
        else:
            channel_to_load = [CHANNEL_PRETRAIN[x] for x in channel_to_load]

        weights = SegformerModel.from_pretrained(self.url).state_dict()
        keys = weights.keys()
        fkeys = [k for k in keys if int(k.split('.')[2]) < self.n_blocks]
        weights = {k: weights[k] for k in fkeys}
        weights['encoder.patch_embeddings.0.proj.weight'] = \
            weights['encoder.patch_embeddings.0.proj.weight'][:, channel_to_load]
        self.encoder.load_state_dict(weights)

    def hug_forward(self, x):
        return self.encoder(x, output_hidden_states=True).hidden_states

    def base_partial_forward(self, x: Tensor, block_slice) -> Tensor:
        B = x.shape[0]

        outputs = ()
        for i in range(self.n_blocks)[block_slice]:
            x, H, W = getattr(self, f"patch_embed{i+1}")(x)
            for blk in getattr(self, f"block{i+1}"):
                x = blk(x, H, W)
            x = getattr(self, f"norm{i+1}")(x).reshape(B, H, W, -1).permute(0, 3, 1, 2)
            outputs = outputs + (x,)
        return outputs

    def base_forward(self, x):
        return self.base_partial_forward(x, slice(self.n_blocks))

    def hug_partial_forward(self, pixel_values, block_slice):
        batch_size = pixel_values.shape[0]
        all_hidden_states = ()
        hidden_states = pixel_values
        output_attentions = False
        for idx, x in list(enumerate(zip(
                self.encoder.encoder.patch_embeddings,
                self.encoder.encoder.block,
                self.encoder.encoder.layer_norm
        )))[block_slice]:
            embedding_layer, block_layer, norm_layer = x
            # first, obtain patch embeddings
            hidden_states, height, width = embedding_layer(hidden_states)
            # second, send embeddings through blocks
            for i, blk in enumerate(block_layer):
                layer_outputs = blk(hidden_states, height, width, output_attentions)
                hidden_states = layer_outputs[0]
            # third, apply layer norm
            hidden_states = norm_layer(hidden_states)
            # fourth, optionally reshape back to (batch_size, num_channels, height, width)
            if idx != len(self.encoder.encoder.patch_embeddings) - 1 or (
                idx == len(self.encoder.encoder.patch_embeddings) - 1 and self.encoder.encoder.config.reshape_last_stage
            ):
                hidden_states = hidden_states.reshape(batch_size, height, width, -1).permute(0, 3, 1, 2).contiguous()
            all_hidden_states = all_hidden_states + (hidden_states,)
        return all_hidden_states


class FusionBlock(nn.Module):
    def __init__(self, channels, fusion_type="conv_sum", p_local=0.1, p_glob=0.5):
        super().__init__()
        self.conv1 = ConvModule(channels, channels, k=1, p=0)
        self.conv2 = ConvModule(channels, channels, k=1, p=0)
        if fusion_type == "conv_sum_drop":
            self.multi_drop = MultiDropPath(num_inputs=2, p=p_glob)
            self.drop = DropPath(p_local)
            self.forward = self.drop_forward
        else:
            self.forward = self.base_forward

    def base_forward(self, x1, x2):
        return self.conv1(x1) + self.conv2(x2)

    def drop_forward(self, x1, x2):
        y1 = x1 + self.drop(self.conv1(x1))
        y2 = x2 + self.drop(self.conv2(x2))
        return sum(self.multi_drop([y1, y2]))


class MiTFusion(nn.Module):
    def __init__(self, channel_dims: list, p_local=0.1, p_glob=0.5, fusion_type="conv_sum"):
        super().__init__()
        self.channel_dims = channel_dims
        for i in range(len(channel_dims)):
            setattr(self, f'fusion_{i}', FusionBlock(channel_dims[i],
                                                     fusion_type=fusion_type,
                                                     p_local=p_local,
                                                     p_glob=p_glob))

    def forward(self, x: Tuple[List]) -> List:
        y = [getattr(self, f"fusion_{i}")(f1, f2)
             for i, (f1, f2) in enumerate(zip(*x))]
        return y


if __name__ == '__main__':
    model = MiT('B0')
    x = torch.zeros(1, 3, 224, 224)
    outs = model(x)
    for y in outs:
        print(y.shape)



import torch 
import math
from torch import nn, Tensor
from torch.nn import functional as F
from ezdl.models.backbones import *
from ezdl.models.layers import ConvModule


class SpatialPath(nn.Module):
    def __init__(self, c1, c2) -> None:
        super().__init__()
        ch = 64
        self.conv_7x7 = ConvModule(c1, ch, 7, 2, 3)
        self.conv_3x3_1 = ConvModule(ch, ch, 3, 2, 1)
        self.conv_3x3_2 = ConvModule(ch, ch, 3, 2, 1)
        self.conv_1x1 = ConvModule(ch, c2, 1, 1, 0)

    def forward(self, x):           
        x = self.conv_7x7(x)
        x = self.conv_3x3_1(x)
        x = self.conv_3x3_2(x)
        return self.conv_1x1(x)


class ContextPath(nn.Module):
    def __init__(self, backbone: nn.Module) -> None:
        super().__init__()
        self.backbone = backbone
        c3, c4 = self.backbone.channels[-2:]

        self.arm16 = AttentionRefinmentModule(c3, 128)
        self.arm32 = AttentionRefinmentModule(c4, 128)

        self.global_context = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            ConvModule(c4, 128, 1, 1, 0)
        )

        self.up16 = nn.Upsample(scale_factor=2.0, mode='bilinear', align_corners=True)
        self.up32 = nn.Upsample(scale_factor=2.0, mode='bilinear', align_corners=True)

        self.refine16 = ConvModule(128, 128, 3, 1, 1)
        self.refine32 = ConvModule(128, 128, 3, 1, 1)


    def forward(self, x):
        _, _, down16, down32 = self.backbone(x)                 # 4x256x64x128, 4x512x32x64

        arm_down16 = self.arm16(down16)                 # 4x128x64x128
        arm_down32 = self.arm32(down32)                 # 4x128x32x64

        global_down32 = self.global_context(down32)     # 4x128x1x1
        global_down32 = F.interpolate(global_down32, size=down32.size()[2:], mode='bilinear', align_corners=True)   # 4x128x32x64

        arm_down32 = arm_down32 + global_down32                         # 4x128x32x64
        arm_down32 = self.up32(arm_down32)                  # 4x128x64x128
        arm_down32 = self.refine32(arm_down32)              # 4x128x64x128

        arm_down16 = arm_down16 + arm_down32                            # 4x128x64x128
        arm_down16 = self.up16(arm_down16)                  # 4x128x128x256
        arm_down16 = self.refine16(arm_down16)              # 4x128x128x256      

        return arm_down16, arm_down32


class AttentionRefinmentModule(nn.Module):
    def __init__(self, c1, c2) -> None:
        super().__init__()
        self.conv_3x3 = ConvModule(c1, c2, 3, 1, 1)

        self.attention = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(c2, c2, 1, bias=False),
            nn.BatchNorm2d(c2),
            nn.Sigmoid()
        )

    def forward(self, x):
        fm = self.conv_3x3(x)
        fm_se = self.attention(fm)
        return fm * fm_se


class FeatureFusionModule(nn.Module):
    def __init__(self, c1, c2, reduction=1) -> None:
        super().__init__()
        self.conv_1x1 = ConvModule(c1, c2, 1, 1, 0)

        self.attention = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(c2, c2 // reduction, 1, bias=False),
            nn.ReLU(True),
            nn.Conv2d(c2 // reduction, c2, 1, bias=False),
            nn.Sigmoid()
        )

    def forward(self, x1, x2):
        fm = torch.cat([x1, x2], dim=1)
        fm = self.conv_1x1(fm)
        fm_se = self.attention(fm)
        return fm + fm * fm_se


class Head(nn.Module):
    def __init__(self, c1, n_classes, upscale_factor, is_aux=False) -> None:
        super().__init__()
        ch = 256 if is_aux else 64
        c2 = n_classes * upscale_factor * upscale_factor
        self.conv_3x3 = ConvModule(c1, ch, 3, 1, 1)
        self.conv_1x1 = nn.Conv2d(ch, c2, 1, 1, 0)
        self.upscale = nn.PixelShuffle(upscale_factor)

    def forward(self, x):
        x = self.conv_1x1(self.conv_3x3(x))
        return self.upscale(x)


class BiSeNetv1(nn.Module):
    def __init__(self, backbone: str = 'ResNet-18', num_classes: int = 19) -> None:
        super().__init__()
        backbone, variant = backbone.split('-')

        self.context_path = ContextPath(eval(backbone)(variant))
        self.spatial_path = SpatialPath(3, 128)
        self.ffm = FeatureFusionModule(256, 256)

        self.output_head = Head(256, num_classes, upscale_factor=8, is_aux=False)
        self.context16_head = Head(128, num_classes, upscale_factor=8, is_aux=True)
        self.context32_head = Head(128, num_classes, upscale_factor=16, is_aux=True)

        self.apply(self._init_weights)

    def _init_weights(self, m: nn.Module) -> None:
        if isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out // m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, (nn.LayerNorm, nn.BatchNorm2d)):
            nn.init.ones_(m.weight)
            nn.init.zeros_(m.bias)

    def init_pretrained(self, pretrained: str = None) -> None:
        if pretrained:
            self.context_path.backbone.load_state_dict(torch.load(pretrained, map_location='cpu'), strict=False)

    def forward(self, x):                                       # 4x3x1024x2048           
        spatial_out = self.spatial_path(x)                      # 4x128x128x256
        context16, context32 = self.context_path(x)             # 4x128x128x256, 4x128x64x128

        fm_fuse = self.ffm(spatial_out, context16)              # 4x256x128x256
        output = self.output_head(fm_fuse)                      # 4xn_classesx1024x2048

        if self.training:
            context_out16 = self.context16_head(context16)      # 4xn_classesx1024x2048
            context_out32 = self.context32_head(context32)      # 4xn_classesx1024x2048
            return output, context_out16, context_out32
        return output    


if __name__ == '__main__':
    model = BiSeNetv1('MobileNetV2-1.0', 19)
    # model.init_pretrained('checkpoints/backbones/resnet/resnet18.pth')
    model.eval()
    image = torch.randn(1, 3, 224, 224)
    output = model(image)
    print(output.shape)


import numpy as np
import torchvision.models as models
import torch
from ptflops import get_model_complexity_info
from models import MODELS


def seg_model_flops(model, size, verbose=False, per_layer_stats=False, model_args={}):
    n_channels, w, h = size
    net = MODELS[model]({'input_channels': n_channels, 'num_classes': 3, "output_channels": 3, **model_args})
    macs, params = get_model_complexity_info(net, (n_channels, w, h), as_strings=True,
                                             print_per_layer_stat=per_layer_stats, verbose=verbose)
    print('{:<30}  {:<8}'.format('Computational complexity: ', macs))
    print('{:<30}  {:<8}'.format('Number of parameters: ', params))


def seg_inference_throughput(model, size, batch_size, device, model_args={}):
    n_channels, w, h = size
    net = MODELS[model]({'input_channels': n_channels, 'num_classes': 3, "output_channels": 3, **model_args}).to(device)
    dummy_input = torch.randn(batch_size, n_channels, w, h, dtype=torch.float).to(device)
    repetitions = 100
    warmup = 50
    total_time = 0
    with torch.no_grad():
        for rep in range(warmup):
            _ = net(dummy_input)
        for rep in range(repetitions):
            starter, ender = torch.cuda.Event(enable_timing=True), torch.cuda.Event(enable_timing=True)
            starter.record()
            _ = net(dummy_input)
            ender.record()
            torch.cuda.synchronize()
            curr_time = starter.elapsed_time(ender) / 1000
            total_time += curr_time
    throughput = (repetitions * batch_size) / total_time
    print('Final Throughput:', throughput)


def seg_inference_inference_per_second(model, size, batch_size, device, model_args={}):
    n_channels, w, h = size
    net = MODELS[model]({'input_channels': n_channels, 'num_classes': 3, "output_channels": 3, **model_args}).to(device)
    dummy_input = torch.randn(batch_size, n_channels, w, h, dtype=torch.float).to(device)
    starter, ender = torch.cuda.Event(enable_timing=True), torch.cuda.Event(enable_timing=True)
    repetitions = 500
    timings = np.zeros((repetitions, 1))
    # GPU-WARM-UP
    for _ in range(50):
        _ = net(dummy_input)
    # MEASURE PERFORMANCE
    with torch.no_grad():
        for rep in range(repetitions):
            starter.record()
            _ = net(dummy_input)
            ender.record()
            # WAIT FOR GPU SYNC
            torch.cuda.synchronize()
            curr_time = starter.elapsed_time(ender)
            timings[rep] = curr_time
    mean_syn = np.sum(timings) / repetitions
    std_syn = np.std(timings)
    print(f"Mean inference time: {mean_syn} ms")
    print(f"Time per example {mean_syn / batch_size} ms")


if __name__ == '__main__':
    models = [
        # ('lawin', 1, {'backbone_pretrained': True}),
        # ('lawin', 2, {'backbone_pretrained': True}),
        # ('lawin', 3, {'backbone_pretrained': True}),
        # ('lawin', 4, {'backbone_pretrained': True, 'main_pretrained': ['R', 'G', 'G', 'G']}),
        #
        # ('lawin', 1, {'backbone': 'MiT-B1'}),
        # ('lawin', 2, {'backbone': 'MiT-B1'}),
        ('lawin', 3, {'backbone': 'MiT-B1'}),
        # ('lawin', 4, {'backbone': 'MiT-B1'}),
        #
        # ('laweed', 1, {'backbone_pretrained': True}),
        # ('laweed', 2, {'backbone_pretrained': True}),
        # ('laweed', 3, {'backbone_pretrained': True}),
        # ('laweed', 4, {'backbone_pretrained': True, 'main_pretrained': ['R', 'G', 'G', 'G']}),
        #
        # ('laweed', 1, {'backbone_pretrained': True, 'backbone': 'MiT-B1'}),
        # ('laweed', 2, {'backbone_pretrained': True, 'backbone': 'MiT-B1'}),
        # ('laweed', 3, {'backbone_pretrained': True, 'backbone': 'MiT-B1'}),
        # ('laweed', 4, {'backbone_pretrained': True, 'main_pretrained': ['R', 'G', 'G', 'G'], 'backbone': 'MiT-B1'}),
        #
        #
        # ('splitlawin', 3, {'main_channels': 2}),
        # ('splitlawin', 4, {'main_channels': 2}),
        # ('splitlawin', 3, {'main_channels': 2, 'backbone': 'MiT-B1'}),
        # ('splitlawin', 4, {'main_channels': 2, 'backbone': 'MiT-B1'}),
        #
        # ('splitlaweed', 3, {'main_channels': 2, 'main_pretrained': ['R', 'G'], 'side_pretrained': 'G'}),
        # ('splitlaweed', 4, {'main_channels': 2, 'main_pretrained': ['R', 'G'], 'side_pretrained': 'G'}),
        #
        # ('doublelawin', 3, {'main_channels': 2}),
        # ('doublelawin', 4, {'main_channels': 2}),
        # ('doublelawin', 3, {'main_channels': 2, 'backbone': 'MiT-B1'}),
        # ('doublelawin', 4, {'main_channels': 2, 'backbone': 'MiT-B1'}),
        #
        # ('doublelaweed', 3, {'main_channels': 2, 'main_pretrained': ['R', 'G'], 'side_pretrained': 'G'}),
        # ('doublelaweed', 4, {'main_channels': 2, 'main_pretrained': ['R', 'G'], 'side_pretrained': 'G'}),
        # ('segnet', 1, {}),
        # ('segnet', 2, {}),
        # ('segnet', 3, {}),
        # ('segnet', 4, {}),
        ('resnet', 3, {"model_name": "50"}),
    ]
    per_layer_stats = False
    verbose = False
    batch_size = 32
    wh = (512, 512)
    for model, channels, args in models:
        with torch.cuda.device(0):
            size = (channels, ) + wh
            print(f"Model: {model}")
            print(f"Size: {size}")
            # seg_model_flops(model, channels, verbose, per_layer_stats, args)
            # seg_inference_throughput(model, channels, batch_size, 'cuda', args)
            seg_inference_inference_per_second(model, size, batch_size, 'cuda', args)
            torch.cuda.empty_cache()
            print()


import torch
from torch import Tensor
from torch.nn import functional as F
from ezdl.models.base import BaseModel
from ezdl.models.heads import SFHead


class SFNet(BaseModel):
    def __init__(self, backbone: str = 'ResNetD-18', num_classes: int = 19):
        assert 'ResNet' in backbone
        super().__init__(backbone, num_classes)
        self.head = SFHead(self.backbone.channels, 128 if '18' in backbone else 256, num_classes)
        self.apply(self._init_weights)

    def forward(self, x: Tensor) -> Tensor:
        outs = self.backbone(x)
        out = self.head(outs)
        out = F.interpolate(out, size=x.shape[-2:], mode='bilinear', align_corners=True)
        return out


if __name__ == '__main__':
    model = SFNet('ResNetD-18')
    model.init_pretrained('checkpoints/backbones/resnetd/resnetd18.pth')
    x = torch.randn(2, 3, 224, 224)
    y = model(x)
    print(y.shape)


import torch
from torch import nn, Tensor
from ezdl.models.layers import DropPath


class LayerNorm(nn.Module):
    """Channel first layer norm
    """
    def __init__(self, normalized_shape, eps=1e-6) -> None:
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.eps = eps

    def forward(self, x: Tensor) -> Tensor:
        u = x.mean(1, keepdim=True)
        s = (x - u).pow(2).mean(1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.eps)
        x = self.weight[:, None, None] * x + self.bias[:, None, None]
        return x

    
class Block(nn.Module):
    def __init__(self, dim, dpr=0., init_value=1e-6):
        super().__init__()
        self.dwconv = nn.Conv2d(dim, dim, 7, 1, 3, groups=dim)
        self.norm = nn.LayerNorm(dim, eps=1e-6)
        self.pwconv1 = nn.Linear(dim, 4*dim)
        self.act = nn.GELU()
        self.pwconv2 = nn.Linear(4*dim, dim)
        self.gamma = nn.Parameter(init_value * torch.ones((dim)), requires_grad=True) if init_value > 0 else None
        self.drop_path = DropPath(dpr) if dpr > 0. else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        input = x
        x = self.dwconv(x)
        x = x.permute(0, 2, 3, 1)   # NCHW to NHWC
        x = self.norm(x)
        x = self.pwconv1(x)
        x = self.act(x)
        x = self.pwconv2(x)

        if self.gamma is not None:
            x = self.gamma * x
        
        x = x.permute(0, 3, 1, 2)
        x = input + self.drop_path(x)
        return x


class Stem(nn.Sequential):
    def __init__(self, c1, c2, k, s):
        super().__init__(
            nn.Conv2d(c1, c2, k, s),
            LayerNorm(c2)
        )


class Downsample(nn.Sequential):
    def __init__(self, c1, c2, k, s):
        super().__init__(
            LayerNorm(c1),
            nn.Conv2d(c1, c2, k, s)
        )


convnext_settings = {
    'T': [[3, 3, 9, 3], [96, 192, 384, 768], 0.0],       # [depths, dims, dpr]
    'S': [[3, 3, 27, 3], [96, 192, 384, 768], 0.0],
    'B': [[3, 3, 27, 3], [128, 256, 512, 1024], 0.0]
}


class ConvNeXt(nn.Module):     
    def __init__(self, model_name: str = 'T') -> None:
        super().__init__()
        assert model_name in convnext_settings.keys(), f"ConvNeXt model name should be in {list(convnext_settings.keys())}"
        depths, embed_dims, drop_path_rate = convnext_settings[model_name]
        self.channels = embed_dims
    
        self.downsample_layers = nn.ModuleList([
            Stem(3, embed_dims[0], 4, 4),
            *[Downsample(embed_dims[i], embed_dims[i+1], 2, 2) for i in range(3)]
        ])

        self.stages = nn.ModuleList()
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]
        cur = 0

        for i in range(4):
            stage = nn.Sequential(*[
                Block(embed_dims[i], dpr[cur+j])
            for j in range(depths[i])])
            self.stages.append(stage)
            cur += depths[i]

        for i in range(4):
            self.add_module(f"norm{i}", LayerNorm(embed_dims[i]))

    def forward(self, x: Tensor):
        outs = []

        for i in range(4):
            x = self.downsample_layers[i](x)
            x = self.stages[i](x)
            norm_layer = getattr(self, f"norm{i}")
            outs.append(norm_layer(x))
        return outs


if __name__ == '__main__':
    model = ConvNeXt('T')
    # model.load_state_dict(torch.load('C:\\Users\\sithu\\Documents\\weights\\backbones\\convnext\\convnext_tiny_1k_224_ema.pth', map_location='cpu')['model'], strict=False)
    x = torch.randn(1, 3, 224, 224)
    feats = model(x)
    for y in feats:
        print(y.shape)


"""
Pytorch implementation of SegNet (https://arxiv.org/pdf/1511.00561.pdf)
"""

from __future__ import print_function
import torch.nn as nn

F = nn.functional
DEBUG = False

vgg16_dims = [
    (64, 64, 'M'),  # Stage - 1
    (128, 128, 'M'),  # Stage - 2
    (256, 256, 256, 'M'),  # Stage - 3
    (512, 512, 512, 'M'),  # Stage - 4
    (512, 512, 512, 'M')  # Stage - 5
]

decoder_dims = [
    ('U', 512, 512, 512),  # Stage - 5
    ('U', 512, 512, 512),  # Stage - 4
    ('U', 256, 256, 256),  # Stage - 3
    ('U', 128, 128),  # Stage - 2
    ('U', 64, 64)  # Stage - 1
]


class SegNet(nn.Module):

    def __init__(self, arch_params):
        super(SegNet, self).__init__()
        in_chn, out_chn = arch_params['input_channels'], arch_params['output_channels']
        # SegNet Architecture
        # Takes input of size in_chn = 3 (RGB images have 3 channels)
        # Outputs size label_chn (N # of classes)

        # ENCODING consists of 5 stages
        # Stage 1, 2 has 2 layers of Convolution + Batch Normalization + Max Pool respectively
        # Stage 3, 4, 5 has 3 layers of Convolution + Batch Normalization + Max Pool respectively

        # General Max Pool 2D for ENCODING layers
        # Pooling indices are stored for Upsampling in DECODING layers

        self.in_chn = in_chn
        self.out_chn = out_chn

        self.MaxEn = nn.MaxPool2d(2, stride=2, return_indices=True)

        self.ConvEn11 = nn.Conv2d(self.in_chn, 64, kernel_size=3, padding=1)
        self.BNEn11 = nn.BatchNorm2d(64,)
        self.ConvEn12 = nn.Conv2d(64, 64, kernel_size=3, padding=1)
        self.BNEn12 = nn.BatchNorm2d(64,)

        self.ConvEn21 = nn.Conv2d(64, 128, kernel_size=3, padding=1)
        self.BNEn21 = nn.BatchNorm2d(128,)
        self.ConvEn22 = nn.Conv2d(128, 128, kernel_size=3, padding=1)
        self.BNEn22 = nn.BatchNorm2d(128,)

        self.ConvEn31 = nn.Conv2d(128, 256, kernel_size=3, padding=1)
        self.BNEn31 = nn.BatchNorm2d(256,)
        self.ConvEn32 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.BNEn32 = nn.BatchNorm2d(256,)
        self.ConvEn33 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.BNEn33 = nn.BatchNorm2d(256,)

        self.ConvEn41 = nn.Conv2d(256, 512, kernel_size=3, padding=1)
        self.BNEn41 = nn.BatchNorm2d(512,)
        self.ConvEn42 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.BNEn42 = nn.BatchNorm2d(512,)
        self.ConvEn43 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.BNEn43 = nn.BatchNorm2d(512,)

        self.ConvEn51 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.BNEn51 = nn.BatchNorm2d(512,)
        self.ConvEn52 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.BNEn52 = nn.BatchNorm2d(512,)
        self.ConvEn53 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.BNEn53 = nn.BatchNorm2d(512,)

        # DECODING consists of 5 stages
        # Each stage corresponds to their respective counterparts in ENCODING

        # General Max Pool 2D/Upsampling for DECODING layers
        self.MaxDe = nn.MaxUnpool2d(2, stride=2)

        self.ConvDe53 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.BNDe53 = nn.BatchNorm2d(512,)
        self.ConvDe52 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.BNDe52 = nn.BatchNorm2d(512,)
        self.ConvDe51 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.BNDe51 = nn.BatchNorm2d(512,)

        self.ConvDe43 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.BNDe43 = nn.BatchNorm2d(512,)
        self.ConvDe42 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.BNDe42 = nn.BatchNorm2d(512,)
        self.ConvDe41 = nn.Conv2d(512, 256, kernel_size=3, padding=1)
        self.BNDe41 = nn.BatchNorm2d(256,)

        self.ConvDe33 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.BNDe33 = nn.BatchNorm2d(256,)
        self.ConvDe32 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.BNDe32 = nn.BatchNorm2d(256,)
        self.ConvDe31 = nn.Conv2d(256, 128, kernel_size=3, padding=1)
        self.BNDe31 = nn.BatchNorm2d(128,)

        self.ConvDe22 = nn.Conv2d(128, 128, kernel_size=3, padding=1)
        self.BNDe22 = nn.BatchNorm2d(128,)
        self.ConvDe21 = nn.Conv2d(128, 64, kernel_size=3, padding=1)
        self.BNDe21 = nn.BatchNorm2d(64,)

        self.ConvDe12 = nn.Conv2d(64, 64, kernel_size=3, padding=1)
        self.BNDe12 = nn.BatchNorm2d(64,)
        self.ConvDe11 = nn.Conv2d(64, self.out_chn, kernel_size=3, padding=1)
        self.BNDe11 = nn.BatchNorm2d(self.out_chn,)

    def forward(self, x):
        # ENCODE LAYERS
        # Stage 1
        x = F.relu(self.BNEn11(self.ConvEn11(x)))
        x = F.relu(self.BNEn12(self.ConvEn12(x)))
        x, ind1 = self.MaxEn(x)
        size1 = x.size()

        # Stage 2
        x = F.relu(self.BNEn21(self.ConvEn21(x)))
        x = F.relu(self.BNEn22(self.ConvEn22(x)))
        x, ind2 = self.MaxEn(x)
        size2 = x.size()

        # Stage 3
        x = F.relu(self.BNEn31(self.ConvEn31(x)))
        x = F.relu(self.BNEn32(self.ConvEn32(x)))
        x = F.relu(self.BNEn33(self.ConvEn33(x)))
        x, ind3 = self.MaxEn(x)
        size3 = x.size()

        # Stage 4
        x = F.relu(self.BNEn41(self.ConvEn41(x)))
        x = F.relu(self.BNEn42(self.ConvEn42(x)))
        x = F.relu(self.BNEn43(self.ConvEn43(x)))
        x, ind4 = self.MaxEn(x)
        size4 = x.size()

        # Stage 5
        x = F.relu(self.BNEn51(self.ConvEn51(x)))
        x = F.relu(self.BNEn52(self.ConvEn52(x)))
        x = F.relu(self.BNEn53(self.ConvEn53(x)))
        x, ind5 = self.MaxEn(x)
        size5 = x.size()

        # DECODE LAYERS
        # Stage 5
        x = self.MaxDe(x, ind5, output_size=size4)
        x = F.relu(self.BNDe53(self.ConvDe53(x)))
        x = F.relu(self.BNDe52(self.ConvDe52(x)))
        x = F.relu(self.BNDe51(self.ConvDe51(x)))

        # Stage 4
        x = self.MaxDe(x, ind4, output_size=size3)
        x = F.relu(self.BNDe43(self.ConvDe43(x)))
        x = F.relu(self.BNDe42(self.ConvDe42(x)))
        x = F.relu(self.BNDe41(self.ConvDe41(x)))

        # Stage 3
        x = self.MaxDe(x, ind3, output_size=size2)
        x = F.relu(self.BNDe33(self.ConvDe33(x)))
        x = F.relu(self.BNDe32(self.ConvDe32(x)))
        x = F.relu(self.BNDe31(self.ConvDe31(x)))

        # Stage 2
        x = self.MaxDe(x, ind2, output_size=size1)
        x = F.relu(self.BNDe22(self.ConvDe22(x)))
        x = F.relu(self.BNDe21(self.ConvDe21(x)))

        # Stage 1
        x = self.MaxDe(x, ind1)
        x = F.relu(self.BNDe12(self.ConvDe12(x)))
        x = self.ConvDe11(x)

        return x



import os

import wandb

from ezdl.data.sequoia import WeedMapDatasetInterface
from ezdl.experiment.run import train, Run
from ezdl.experiment.parameters import parse_params
from ezdl.learning.seg_trainer import SegmentationTrainer
from ezdl.utils.utilities import values_to_number, nested_dict_update

from super_gradients.common.abstractions.abstract_logger import get_logger

logger = get_logger(__name__)


def resume_set_of_runs(settings, post_filters=None):
    queries = settings['runs']
    path = settings['path']
    for query in queries:
        filters = query['filters']
        stage = query['stage']
        updated_config = query['updated_config']
        api = wandb.Api()
        runs = api.runs(path=path, filters=filters)
        runs = list(filter(post_filters, runs))
        print("Runs to resume:")
        for run in runs:
            print(f"{run.group} \t - \t {run.name}")
        if len(runs) == 0:
            logger.error(f'No runs found for query {filters} with post_filters: {post_filters}')
        for run in runs:
            resume_run(run, updated_config, stage)


def complete_incompleted_runs(settings):
    print("Going on to complete runs!")
    resume_set_of_runs(settings, lambda x: 'f1' not in x.summary)


def resume_run(wandb_run, updated_config, stage):
    to_resume_run = Run()
    to_resume_run.resume(wandb_run=wandb_run, updated_config=updated_config, phases=stage)
    to_resume_run.launch()


def get_interrupted_run(input_settings):
    namespace = input_settings["name"]
    group = input_settings["group"]
    last_run = wandb.Api().runs(path=namespace, filters={"group": group,}, order="-created_at")[0]
    filters = {"group": group, "name": last_run.id}
    stage = ["train", "test"]
    updated_config = None
    api = wandb.Api()
    runs = api.runs(path=namespace, filters=filters)
    if len(runs) == 0:
        raise RuntimeError("No runs found")
    if len(runs) > 1:
        raise EnvironmentError("More than 1 run???")
    to_resume_run = Run()
    to_resume_run.resume(wandb_run=runs[0], updated_config=updated_config, phases=stage)
    return to_resume_run


def retrieve_run_to_resume(settings, grids):
    grid_list = [(i, j) for i in range(len(grids)) for j in range(len(grids[i]))]
    dir = settings['tracking_dir']
    exp_log = ExpLog(track_dir=dir, mode="r")
    i, j, status = exp_log.get_last_run()
    index = grid_list.index((i, j))
    try:
        start_grid, start_run = grid_list[index + 1]  # Skip interrupted run
    except IndexError as e:
        if status == "finished \n":
            logger.info(e)
            raise ValueError('No experiment to resume!!')
        else:
            return len(grids), None, True
    resume_last = True if status != "finished \n" else False
    return start_grid, start_run, resume_last


class ExpLog:
    EXP_END = '---\n'
    FINISHED = 'finished \n'
    CRASHED = 'crashed \n'

    def __init__(self, track_dir, mode='a'):
        self.exp_log = open(os.path.join(track_dir if track_dir is not None else '', 'exp_log.txt'), mode)

    def start(self):
        self.exp_log.write(self.EXP_END)
        self.exp_log.flush()

    def write(self, s):
        self.exp_log.write(s)
        self.exp_log.flush()

    def close(self):
        self.exp_log.close()

    def insert_run(self, i, j):
        self.exp_log.write(f'{i} {j},')
        self.exp_log.flush()

    def finish_run(self):
        self.exp_log.write(self.FINISHED)
        self.exp_log.flush()

    def crash_run(self):
        self.exp_log.write(self.CRASHED)
        self.exp_log.flush()

    def get_last_run(self):
        lines = self.exp_log.readlines()
        i = 1
        while lines[-i] in [self.EXP_END, '\n']:
            i += 1
        last_ran = lines[-i]

        code, status = last_ran.split(",")
        i, j = map(int, code.split(" "))
        return i, j, status


import argparse
import json

from ezdl.utils.utilities import update_collection, load_yaml

parser = argparse.ArgumentParser(description='Train and test models')
parser.add_argument('action',
                    help='Choose the action to perform: experiment, resume, resume_run, complete, preprocess, manipulate, app',
                    default="experiment", type=str)
parser.add_argument('--resume', required=False, action='store_true',
                    help='Resume the experiment', default=False)
parser.add_argument('-d', '--dir', required=False, type=str,
                    help='Set the local tracking directory', default=None)
parser.add_argument('-f', '--file', required=False, type=str,
                    help='Set the config file', default=None)
parser.add_argument("--grid", type=int, help="Select the first grid to start from")
parser.add_argument("--run", type=int, help="Select the run in grid to start from")

parser.add_argument("--filters", type=json.loads, help="Filters to query in the resuming mode")
parser.add_argument('-s', "--stage", type=json.loads, help="Stages to execute in the resuming mode")
parser.add_argument('-p', "--path", type=str, help="Path to the tracking url in the resuming mode")

parser.add_argument('--subset', type=str, help="Subset chosen for preprocessing")

if __name__ == '__main__':
    args = parser.parse_args()

    track_dir = args.dir
    filters = args.filters
    stage = args.stage
    path = args.path
    action = args.action
    file = args.file

    if action == "resume_run":
        from ezdl.experiment.resume import resume_set_of_runs
        param_path = args.file or 'resume.yaml'
        settings = load_yaml(param_path)
        settings['runs'][0]['filters'] = update_collection(settings['runs'][0]['filters'], filters)
        settings['runs'][0]['stage'] = update_collection(settings['runs'][0]['stage'], stage)
        settings = update_collection(settings, path, key="path")
        resume_set_of_runs(settings)
    elif action == 'experiment':
        from ezdl.experiment.experiment import experiment
        param_path = args.file or 'parameters.yaml'
        settings = load_yaml(param_path)
        settings['experiment'] = update_collection(settings['experiment'], args.resume, key='resume')
        settings['experiment'] = update_collection(settings['experiment'], args.grid, key='start_from_grid')
        settings['experiment'] = update_collection(settings['experiment'], args.run, key='start_from_run')
        settings['experiment'] = update_collection(settings['experiment'], track_dir, key='tracking_dir')
        experiment(settings)
    elif action == 'preprocess':
        from ezdl.preprocessing import preprocess
        preprocess(subset=args.subset)
    elif action == 'complete':
        from ezdl.experiment.resume import complete_incompleted_runs
        param_path = args.file or 'resume.yaml'
        settings = load_yaml(param_path)
        complete_incompleted_runs(settings)
    elif action == 'manipulation':
        from ezdl.wandb_manip import manipulate
        manipulate()
    elif action in ['app', 'webapp', 'frontend']:
        from ezdl.app import frontend
        frontend()
    else:
        raise ValueError("Action not recognized")



import torch 
from torch import nn, Tensor


class HSigmoid(nn.Module):
    def __init__(self):
        super().__init__()
        self.relu = nn.ReLU6(True)

    def forward(self, x: Tensor) -> Tensor:
        return self.relu(x + 3) / 6


class HSwish(nn.Module):
    def __init__(self):
        super().__init__()
        self.sigmoid = HSigmoid()

    def forward(self, x: Tensor) -> Tensor:
        return x * self.sigmoid(x)


class ChannelShuffle(nn.Module):
    def __init__(self, groups):
        super(ChannelShuffle, self).__init__()
        self.groups = groups

    def forward(self, x):
        b, c, h, w = x.size()

        channels_per_group = c // self.groups

        # reshape
        x = x.view(b, self.groups, channels_per_group, h, w)

        x = torch.transpose(x, 1, 2).contiguous()
        out = x.view(b, -1, h, w)

        return out


def _make_divisible(v, divisor, min_value=None):
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


class DYShiftMax(nn.Module):
    def __init__(self, c1, c2, 
        init_a=[0.0, 0.0],
        init_b=[0.0, 0.0],
        act_relu=True,
        g=None,
        reduction=4,
        expansion=False
    ):
        super().__init__()
        self.exp = 4 if act_relu else 2
        self.init_a = init_a
        self.init_b = init_b
        self.c2 = c2

        self.avg_pool = nn.Sequential(
            nn.Sequential(),
            nn.AdaptiveAvgPool2d(1)
        )
        
        squeeze = _make_divisible(c1 // reduction, 4)

        self.fc = nn.Sequential(
            nn.Linear(c1, squeeze),
            nn.ReLU(True),
            nn.Linear(squeeze, c2 * self.exp),
            HSigmoid()
        )

        g = g[1]
        if g != 1 and expansion:
            g = c1 // g
        
        gc = c1 // g
        index = torch.Tensor(range(c1)).view(1, c1, 1, 1)
        index = index.view(1, g, gc, 1, 1)
        indexgs = torch.split(index, [1, g-1], dim=1)
        indexgs = torch.cat([indexgs[1], indexgs[0]], dim=1)
        indexs = torch.split(indexgs, [1, gc-1], dim=2)
        indexs = torch.cat([indexs[1], indexs[0]], dim=2)
        self.index = indexs.view(c1).long()
        
        
    def forward(self, x: Tensor) -> Tensor:
        B, C, H, W = x.shape
        x_out = x
        
        y = self.avg_pool(x).view(B, C)
        y = self.fc(y).view(B, -1, 1, 1)
        y = (y - 0.5) * 4.0

        x2 = x_out[:, self.index, :, :]

        if self.exp == 4:
            a1, b1, a2, b2 = torch.split(y, self.c2, dim=1)

            a1 = a1 + self.init_a[0]
            a2 = a2 + self.init_b[1]
            b1 = b1 + self.init_b[0]
            b2 = b2 + self.init_b[1]

            z1 = x_out * a1 + x2 * b1
            z2 = x_out * a2 + x2 * b2

            out = torch.max(z1, z2)

        elif self.exp == 2:
            a1, b1 = torch.split(y, self.c2, dim=1)
            a1 = a1 + self.init_a[0]
            b1 = b1 + self.init_b[0]
            out = x_out * a1 + x2 * b1

        return out


class SwishLinear(nn.Module):
    def __init__(self, c1, c2):
        super().__init__()
        self.linear = nn.Sequential(
            nn.Linear(c1, c2),
            nn.BatchNorm1d(c2),
            HSwish()
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.linear(x)


class SpatialSepConvSF(nn.Module):
    def __init__(self, c1, outs, k, s):
        super().__init__()
        o1, o2 = outs
        self.conv = nn.Sequential(
            nn.Conv2d(c1, o1, (k, 1), (s, 1), (k//2, 0), bias=False),
            nn.BatchNorm2d(o1),
            nn.Conv2d(o1, o1*o2, (1, k), (1, s), (0, k//2), groups=o1, bias=False),
            nn.BatchNorm2d(o1*o2),
            ChannelShuffle(o1)
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.conv(x)


class Stem(nn.Module):
    def __init__(self, c1, c2, s, g=(4, 4)):
        super().__init__()
        self.stem = nn.Sequential(
            SpatialSepConvSF(c1, g, 3, s),
            nn.ReLU6(True)
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.stem(x)


class DepthSpatialSepConv(nn.Module):
    def __init__(self, c1, expand, k, s):
        super().__init__()
        exp1, exp2 = expand
        ch = c1 * exp1
        c2 = c1 * exp1 * exp2

        self.conv = nn.Sequential(
            nn.Conv2d(c1, ch, (k, 1), (s, 1), (k//2, 0), groups=c1, bias=False),
            nn.BatchNorm2d(ch),
            nn.Conv2d(ch, c2, (1, k), (1, s), (0, k//2), groups=ch, bias=False),
            nn.BatchNorm2d(c2)
        )
    
    def forward(self, x: Tensor) -> Tensor:
        return self.conv(x)


class PWConv(nn.Module):
    def __init__(self, c1, c2, g=2):
        super().__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(c1, c2, 1, 1, 0, groups=g[0], bias=False),
            nn.BatchNorm2d(c2)
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.conv(x)


class MicroBlock(nn.Module):
    def __init__(self, c1, c2, k=3, s=1, 
        t1=(2, 2), 
        gs1=4, 
        groups_1x1=(1, 1), 
        dy=(2, 0, 1),
        r = 1,
        init_a=(1.0, 1.0),
        init_b=(0.0, 0.0),
    ):
        super().__init__()
        self.identity = s == 1 and c1 == c2
        y1, y2, y3 = dy
        _, g1, g2 = groups_1x1
        reduction = 8 * r
        ch2 = c1 * t1[0] * t1[1]

        if gs1[0] == 0:
            self.layers = nn.Sequential(
                DepthSpatialSepConv(c1, t1, k, s),
                DYShiftMax(ch2, ch2, init_a, init_b, True if y2 == 2 else False, gs1, reduction) if y2 > 0 else nn.ReLU6(True),
                ChannelShuffle(gs1[1]),
                ChannelShuffle(ch2 // 2) if y2 != 0 else nn.Sequential(),
                PWConv(ch2, c2, (g1, g2)),
                DYShiftMax(c2, c2, [1.0, 0.0], [0.0, 0.0], False, (g1, g2), reduction//2) if y3 > 0 else nn.Sequential(),
                ChannelShuffle(g2),
                ChannelShuffle(c2 // 2) if c2%2 == 0 and y3 != 0 else nn.Sequential()
            )
        elif g2 == 0:
            self.layers = nn.Sequential(
                PWConv(c1, ch2, gs1),
                DYShiftMax(ch2, ch2, [1.0, 0.0], [0.0, 0.0], False, gs1, reduction) if y3 > 0 else nn.Sequential()
            )
        else:
            self.layers = nn.Sequential(
                PWConv(c1, ch2, gs1),
                DYShiftMax(ch2, ch2, init_a, init_b, True if y1 == 2 else False, gs1, reduction) if y1 > 0 else nn.ReLU6(True),
                ChannelShuffle(gs1[1]),
                DepthSpatialSepConv(ch2, (1, 1), k, s),
                nn.Sequential(),
                DYShiftMax(ch2, ch2, init_a, init_b, True if y2 == 2 else False, gs1, reduction, True) if y2 > 0 else nn.ReLU6(True),
                ChannelShuffle(ch2 // 4) if y1 != 0 and y2 != 0 else nn.Sequential() if y1 == 0 and y2 == 0 else ChannelShuffle(ch2 // 2),
                PWConv(ch2, c2, (g1, g2)),
                DYShiftMax(c2, c2, [1.0, 0.0], [0.0, 0.0], False, (g1, g2), reduction=reduction//2 if c2 < ch2 else reduction) if y3 > 0 else nn.Sequential(),
                ChannelShuffle(g2),
                ChannelShuffle(c2 // 2) if y3 != 0 else nn.Sequential()
            )

    def forward(self, x: Tensor) -> Tensor:
        identity = x
        out = self.layers(x)
        if self.identity:
            out += identity
        return out


micronet_settings = {
    'M1': [
        6,              # stem_ch
        [3, 2],         # stem_groups
        960,            # out_ch
        [1.0, 1.0],     # init_a
        [0.0, 0.0],     # init_b
        [1, 2, 4, 7],   # out indices
        [8, 16, 32, 576],
        [
            #s, c, ks, c1, c2, g1, g2, c3, g3, g4, y1, y2, y3, r
            [2,   8, 3, 2, 2,  0,  6,   8,  2,  2, 2, 0, 1, 1],  #6->12(0, 0)->24  ->8(4,2)->8
            [2,  16, 3, 2, 2,  0,  8,  16,  4,  4, 2, 2, 1, 1], #8->16(0, 0)->32  ->16(4,4)->16
            [2,  16, 5, 2, 2,  0, 16,  16,  4,  4, 2, 2, 1, 1], #16->32(0, 0)->64  ->16(8,2)->16
            [1,  32, 5, 1, 6,  4,  4,  32,  4,  4, 2, 2, 1, 1], #16->16(2,8)->96 ->32(8,4)->32
            [2,  64, 5, 1, 6,  8,  8,  64,  8,  8, 2, 2, 1, 1], #32->32(2,16)->192 ->64(12,4)->64
            [1,  96, 3, 1, 6,  8,  8,  96,  8,  8, 2, 2, 1, 2], #64->64(3,16)->384 ->96(16,6)->96
            [1, 576, 3, 1, 6, 12, 12,   0,  0,  0, 2, 2, 1, 2], #96->96(4,24)->576
        ],
    ],  
    'M2': [
        8, 
        [4, 2],
        1024,
        [1.0, 1.0],
        [0.0, 0.0],
        [1, 3, 6, 9],
        [12, 24, 64, 768],
        [
            #s,  c, ks, c1, c2, g1, g2, c3, g3, g4, y1, y2, y3, r
            [2,  12, 3, 2, 2,  0,  8,  12,  4,  4, 2, 0, 1, 1], #8->16(0, 0)->32  ->12(4,3)->12
            [2,  16, 3, 2, 2,  0, 12,  16,  4,  4, 2, 2, 1, 1], #12->24(0,0)->48  ->16(8, 2)->16
            [1,  24, 3, 2, 2,  0, 16,  24,  4,  4, 2, 2, 1, 1], #16->16(0, 0)->64  ->24(8,3)->24
            [2,  32, 5, 1, 6,  6,  6,  32,  4,  4, 2, 2, 1, 1], #24->24(2, 12)->144  ->32(16,2)->32
            [1,  32, 5, 1, 6,  8,  8,  32,  4,  4, 2, 2, 1, 2], #32->32(2,16)->192 ->32(16,2)->32
            [1,  64, 5, 1, 6,  8,  8,  64,  8,  8, 2, 2, 1, 2], #32->32(2,16)->192 ->64(12,4)->64
            [2,  96, 5, 1, 6,  8,  8,  96,  8,  8, 2, 2, 1, 2], #64->64(4,12)->384 ->96(16,5)->96
            [1, 128, 3, 1, 6, 12, 12, 128,  8,  8, 2, 2, 1, 2], #96->96(5,16)->576->128(16,8)->128
            [1, 768, 3, 1, 6, 16, 16,   0,  0,  0, 2, 2, 1, 2], #128->128(4,32)->768
        ],
    ],
    'M3': [
        12, 
        [4, 3],
        1024,
        [1.0, 0.5],
        [0.0, 0.5],
        [1, 3, 8, 12],
        [16, 24, 80, 864],
        [
            #s,  c, ks, c1, c2, g1, g2, c3, g3, g4, y1, y2, y3, r
            [2,  16, 3, 2, 2,  0, 12,  16,  4,  4, 0, 2, 0, 1], #12->24(0, 0)->48  ->16(8,2)->16
            [2,  24, 3, 2, 2,  0, 16,  24,  4,  4, 0, 2, 0, 1], #16->32(0, 0)->64  ->24(8,3)->24
            [1,  24, 3, 2, 2,  0, 24,  24,  4,  4, 0, 2, 0, 1], #24->48(0, 0)->96  ->24(8,3)->24
            [2,  32, 5, 1, 6,  6,  6,  32,  4,  4, 0, 2, 0, 1], #24->24(2,12)->144  ->32(16,2)->32
            [1,  32, 5, 1, 6,  8,  8,  32,  4,  4, 0, 2, 0, 2], #32->32(2,16)->192 ->32(16,2)->32
            [1,  64, 5, 1, 6,  8,  8,  48,  8,  8, 0, 2, 0, 2], #32->32(2,16)->192 ->48(12,4)->48
            [1,  80, 5, 1, 6,  8,  8,  80,  8,  8, 0, 2, 0, 2], #48->48(3,16)->288 ->80(16,5)->80
            [1,  80, 5, 1, 6, 10, 10,  80,  8,  8, 0, 2, 0, 2], #80->80(4,20)->480->80(20,4)->80
            [2, 120, 5, 1, 6, 10, 10, 120, 10, 10, 0, 2, 0, 2], #80->80(4,20)->480->128(16,8)->128
            [1, 120, 5, 1, 6, 12, 12, 120, 10, 10, 0, 2, 0, 2], #120->128(4,32)->720->128(32,4)->120
            [1, 144, 3, 1, 6, 12, 12, 144, 12, 12, 0, 2, 0, 2], #120->128(4,32)->720->160(32,5)->144
            [1, 864, 3, 1, 6, 12, 12,   0,  0,  0, 0, 2, 0, 2], #144->144(5,32)->864
        ],
    ]
}


class MicroNet(nn.Module):
    def __init__(self, variant: str = 'M1') -> None:
        super().__init__()
        self.inplanes = 64

        assert variant in micronet_settings.keys(), f"MicroNet model name should be in {list(micronet_settings.keys())}"
        input_channel, stem_groups, _, init_a, init_b, out_indices, channels, cfgs = micronet_settings[variant]
        self.out_indices = out_indices
        self.channels = channels

        self.features = nn.ModuleList([Stem(3, input_channel, 2, stem_groups)])

        for s, c, ks, c1, c2, g1, g2, c3, g3, g4, y1, y2, y3, r in cfgs:
            self.features.append(MicroBlock(input_channel, c, ks, s, (c1, c2), (g1, g2), (c3, g3, g4), (y1, y2, y3), r, init_a, init_b))
            input_channel = c

    def forward(self, x: Tensor) -> Tensor:
        outs = []
        for i, m in enumerate(self.features):
            x = m(x)
            if i in self.out_indices:
                outs.append(x)
        return outs


if __name__ == '__main__':
    model = MicroNet('M3')
    # model.load_state_dict(torch.load('checkpoints/backbones/micronet/micronet-m2.pth', map_location='cpu'), strict=False)
    x = torch.zeros(1, 3, 224, 224)
    # outs = model(x)
    # for y in outs:
    #     print(y.shape)
    from fvcore.nn import flop_count_table, FlopCountAnalysis
    print(flop_count_table(FlopCountAnalysis(model, x)))


import torch
from torch import nn, Tensor
from torch.nn import functional as F
from typing import Tuple
from ezdl.models.layers import ConvModule
from ezdl.models.modules import PPM


class UPerHead(nn.Module):
    """Unified Perceptual Parsing for Scene Understanding
    https://arxiv.org/abs/1807.10221
    scales: Pooling scales used in PPM module applied on the last feature
    """
    def __init__(self, in_channels, channel=128, num_classes: int = 19, scales=(1, 2, 3, 6)):
        super().__init__()
        # PPM Module
        self.ppm = PPM(in_channels[-1], channel, scales)

        # FPN Module
        self.fpn_in = nn.ModuleList()
        self.fpn_out = nn.ModuleList()

        for in_ch in in_channels[:-1]: # skip the top layer
            self.fpn_in.append(ConvModule(in_ch, channel, 1))
            self.fpn_out.append(ConvModule(channel, channel, 3, 1, 1))

        self.bottleneck = ConvModule(len(in_channels)*channel, channel, 3, 1, 1)
        self.dropout = nn.Dropout2d(0.1)
        self.conv_seg = nn.Conv2d(channel, num_classes, 1)


    def forward(self, features: Tuple[Tensor, Tensor, Tensor, Tensor]) -> Tensor:
        f = self.ppm(features[-1])
        fpn_features = [f]

        for i in reversed(range(len(features)-1)):
            feature = self.fpn_in[i](features[i])
            f = feature + F.interpolate(f, size=feature.shape[-2:], mode='bilinear', align_corners=False)
            fpn_features.append(self.fpn_out[i](f))

        fpn_features.reverse()
        for i in range(1, len(features)):
            fpn_features[i] = F.interpolate(fpn_features[i], size=fpn_features[0].shape[-2:], mode='bilinear', align_corners=False)
 
        output = self.bottleneck(torch.cat(fpn_features, dim=1))
        output = self.conv_seg(self.dropout(output))
        return output


if __name__ == '__main__':
    model = UPerHead([64, 128, 256, 512], 128)
    x1 = torch.randn(2, 64, 56, 56)
    x2 = torch.randn(2, 128, 28, 28)
    x3 = torch.randn(2, 256, 14, 14)
    x4 = torch.randn(2, 512, 7, 7)
    y = model([x1, x2, x3, x4])
    print(y.shape)


import torch
from torch import nn, Tensor
from torch.nn import functional as F
from ezdl.models.layers import DropPath


class DWConv(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dwconv = nn.Conv2d(dim, dim, 3, 1, 1, groups=dim)

    def forward(self, x: Tensor, H: int, W: int) -> Tensor:
        B, _, C = x.shape
        x = x.transpose(1, 2).view(B, C, H, W)
        x = self.dwconv(x)
        return x.flatten(2).transpose(1, 2)


class MLP(nn.Module):
    def __init__(self, dim, hidden_dim, out_dim=None) -> None:
        super().__init__()
        out_dim = out_dim or dim
        self.fc1 = nn.Linear(dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, out_dim)
        self.dwconv = DWConv(hidden_dim)

    def forward(self, x: Tensor, H: int, W: int) -> Tensor:
        return self.fc2(F.gelu(self.dwconv(self.fc1(x), H, W)))


class Attention(nn.Module):
    def __init__(self, dim, head, sr_ratio):
        super().__init__()
        self.head = head
        self.sr_ratio = sr_ratio 
        self.scale = (dim // head) ** -0.5
        self.q = nn.Linear(dim, dim, bias=True)
        self.kv = nn.Linear(dim, dim*2, bias=True)
        self.proj = nn.Linear(dim, dim)

        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, sr_ratio, sr_ratio)
            self.norm = nn.LayerNorm(dim)

    def forward(self, x: Tensor, H, W) -> Tensor:
        B, N, C = x.shape
        q = self.q(x).reshape(B, N, self.head, C // self.head).permute(0, 2, 1, 3)

        if self.sr_ratio > 1:
            x = x.permute(0, 2, 1).reshape(B, C, H, W)
            x = self.sr(x).reshape(B, C, -1).permute(0, 2, 1)
            x = self.norm(x)
        k, v = self.kv(x).reshape(B, -1, 2, self.head, C // self.head).permute(2, 0, 3, 1, 4)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        return x


class Block(nn.Module):
    def __init__(self, dim, head, sr_ratio=1, mlp_ratio=4, dpr=0.):
        super().__init__()
        self.norm1 = nn.LayerNorm(dim)
        self.attn = Attention(dim, head, sr_ratio)
        self.drop_path = DropPath(dpr) if dpr > 0. else nn.Identity()
        self.norm2 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, int(dim*mlp_ratio))

    def forward(self, x: Tensor, H, W) -> Tensor:
        x = x + self.drop_path(self.attn(self.norm1(x), H, W))
        x = x + self.drop_path(self.mlp(self.norm2(x), H, W))
        return x


class PatchEmbed(nn.Module):
    def __init__(self, c1=3, c2=64, patch_size=7, stride=4):
        super().__init__()
        self.proj = nn.Conv2d(c1, c2, patch_size, stride, patch_size//2)
        self.norm = nn.LayerNorm(c2)

    def forward(self, x: Tensor) -> Tensor:
        x = self.proj(x)
        _, _, H, W = x.shape
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)
        return x, H, W


pvtv2_settings = {
    'B1': [2, 2, 2, 2],    # depths
    'B2': [3, 4, 6, 3],
    'B3': [3, 4, 18, 3],
    'B4': [3, 8, 27, 3],
    'B5': [3, 6, 40, 3]
}


class PVTv2(nn.Module):     
    def __init__(self, model_name: str = 'B1') -> None:
        super().__init__()
        assert model_name in pvtv2_settings.keys(), f"PVTv2 model name should be in {list(pvtv2_settings.keys())}"
        depths = pvtv2_settings[model_name]
        embed_dims = [64, 128, 320, 512]
        drop_path_rate = 0.1
        self.channels = embed_dims
        # patch_embed
        self.patch_embed1 = PatchEmbed(3, embed_dims[0], 7, 4)
        self.patch_embed2 = PatchEmbed(embed_dims[0], embed_dims[1], 3, 2)
        self.patch_embed3 = PatchEmbed(embed_dims[1], embed_dims[2], 3, 2)
        self.patch_embed4 = PatchEmbed(embed_dims[2], embed_dims[3], 3, 2)
        
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]
        # transformer encoder
        cur = 0
        self.block1 = nn.ModuleList([Block(embed_dims[0], 1, 8, 8, dpr[cur+i]) for i in range(depths[0])])
        self.norm1 = nn.LayerNorm(embed_dims[0])

        cur += depths[0]
        self.block2 = nn.ModuleList([Block(embed_dims[1], 2, 4, 8, dpr[cur+i]) for i in range(depths[1])])
        self.norm2 = nn.LayerNorm(embed_dims[1])

        cur += depths[1]
        self.block3 = nn.ModuleList([Block(embed_dims[2], 5, 2, 4, dpr[cur+i]) for i in range(depths[2])])
        self.norm3 = nn.LayerNorm(embed_dims[2])

        cur += depths[2]
        self.block4 = nn.ModuleList([Block(embed_dims[3], 8, 1, 4, dpr[cur+i]) for i in range(depths[3])])
        self.norm4 = nn.LayerNorm(embed_dims[3])

    def forward(self, x: Tensor) -> Tensor:
        B = x.shape[0]
        # stage 1
        x, H, W = self.patch_embed1(x)
        for blk in self.block1:
            x = blk(x, H, W)
        x1 = self.norm1(x).reshape(B, H, W, -1).permute(0, 3, 1, 2)

        # stage 2
        x, H, W = self.patch_embed2(x1)
        for blk in self.block2:
            x = blk(x, H, W)
        x2 = self.norm2(x).reshape(B, H, W, -1).permute(0, 3, 1, 2)

        # stage 3
        x, H, W = self.patch_embed3(x2)
        for blk in self.block3:
            x = blk(x, H, W)
        x3 = self.norm3(x).reshape(B, H, W, -1).permute(0, 3, 1, 2)

        # stage 4
        x, H, W = self.patch_embed4(x3)
        for blk in self.block4:
            x = blk(x, H, W)
        x4 = self.norm4(x).reshape(B, H, W, -1).permute(0, 3, 1, 2)

        return x1, x2, x3, x4


if __name__ == '__main__':
    model = PVTv2('B1')
    model.load_state_dict(torch.load('checkpoints/backbones/pvtv2/pvt_v2_b1.pth', map_location='cpu'), strict=False)
    x = torch.zeros(1, 3, 224, 224)
    outs = model(x)
    for y in outs:
        print(y.shape)


import numpy as np
import torch
from PIL import Image

from ezdl.data.sequoia import WeedMapDataset
from tqdm import tqdm

import torchvision.transforms as transforms
import os
import shutil


def delete_empty_imgs(root, channels, tempdir_check=None):
    if tempdir_check:
        shutil.rmtree(tempdir_check, ignore_errors=True)
        os.makedirs(tempdir_check, exist_ok=True)
    trs = transforms.Compose([])
    dataset = WeedMapDataset(root, transform=trs, return_mask=True, target_transform=trs)
    counter = 0
    for i in tqdm(range(len(dataset))):
        folder, img_name = dataset.index[i]
        img, (gt, mask) = dataset[i]
        if np.min(np.array(mask)) == 255:
            gt.close()
            mask.close()
            gt_path_color = os.path.join(root, folder, 'groundtruth',
                                         folder + '_' + img_name.split('.')[0] + '_GroundTruth_color.png'
                                         )
            gt_path_imap = os.path.join(root, folder, 'groundtruth',
                                        folder + '_' + img_name.split('.')[0] + '_GroundTruth_iMap.png'
                                        )
            gt_path = os.path.join(root, folder, 'groundtruth',
                                   folder + '_' + img_name)
            mask_path = os.path.join(root, folder, 'mask', img_name)

            if tempdir_check:
                if isinstance(img, torch.Tensor):
                    img = Image.fromarray(img.byte().permute(1, 2, 0).numpy())
                img.save(os.path.join(tempdir_check, img_name))
                shutil.copy(gt_path_color, os.path.join(tempdir_check,
                                                        folder + '_' + img_name.split('.')[0] + '_GroundTruth_color.png'))
            os.remove(gt_path_color)
            os.remove(gt_path_imap)
            os.remove(gt_path)
            os.remove(mask_path)
            for c in channels:
                c_path = os.path.join(root, folder, 'tile', c, img_name)
                os.remove(c_path)
            counter += 1
    print('Removed {} empty images'.format(counter))


def copy_dataset(inpath, outpath):
    shutil.copytree(inpath, outpath, dirs_exist_ok=True)
    print('Dataset copied')


def preprocess(subset):
    if subset == "SEQUOIA":
        # SEQUOIA
        channels = ['CIR', 'G', 'NDVI', 'NIR', 'R', 'RE']
        path = 'dataset/raw/Sequoia'
        outpath = 'dataset/processed/Sequoia'
    elif subset == "REDEDGE":
        # REDEDGE 413 removed
        channels = ['CIR', 'G', 'NDVI', 'NIR', 'R', 'RE', 'B']
        path = 'dataset/raw/RedEdge'
        outpath = 'dataset/processed/RedEdge'
        copy_dataset(path, outpath)
    else:
        raise NotImplementedError()
    delete_empty_imgs(outpath, channels, tempdir_check='tmp')





from tqdm import tqdm
from collections import defaultdict
import numpy as np

from data.sequoia import WeedMapDataset

def count_classes(root, folders):
    channels = ['NDVI']
    index = WeedMapDataset.build_index(
        root,
        macro_folders=folders,
        channels=channels,
    )
    sq = WeedMapDataset(root,
                        transform=lambda x: x,
                        target_transform=lambda x: np.array(x),
                        index=index,
                        channels=channels,
                        return_path=True
                        )
    # loop through images
    class_counts = defaultdict(lambda: 0)
    for input, target, additional in tqdm(sq):
        d, counts = np.unique(target, return_counts=True)
        for v, c in zip(d, counts):
            class_counts[v] += c
    print(class_counts)
    values = np.array(list(class_counts.values()))
    print(values / np.sum(values))


if __name__ == '__main__':
    f = ['005']
    r = "./dataset/processed/Sequoia"
    count_classes(r, f)


import numpy as np
import torch


def tensor_to_segmentation_image(prediction: torch.Tensor, cmap) -> np.array:
    segmented_image = np.ones((*prediction.shape, 3), dtype="uint8")
    for class_value, class_color in cmap.items():
        segmented_image[prediction == class_value] = class_color
    return segmented_image




import gc
import os

from super_gradients.common.abstractions.abstract_logger import get_logger
from super_gradients.training.utils.callbacks import Phase

from ezdl.callbacks import WandbCallback, SegmentationVisualizationCallback
from ezdl.data.sequoia import WeedMapDatasetInterface
from ezdl.experiment.parameters import parse_params
from ezdl.learning.seg_trainer import SegmentationTrainer
from ezdl.utils.utilities import dict_to_yaml_string, values_to_number, nested_dict_update

import ezdl.data

logger = get_logger(__name__)


class Run:
    def __init__(self):
        self.params = None
        self.dataset = None
        self.early_stop = None
        self.dataset_params = None
        self.seg_trainer = None
        self.train_params = None
        self.test_params = None
        self.run_params = None
        self.phases = None

    def init(self, params: dict):
        self.seg_trainer = None
        try:
            self.params = params
            self.phases = params['phases']

            self.train_params, self.test_params, self.dataset_params, self.early_stop = parse_params(params)
            self.run_params = params.get('run_params') or {}

            self.seg_trainer = SegmentationTrainer(experiment_name=params['experiment']['group'],
                                                   ckpt_root_dir=params['experiment']['tracking_dir']
                                                   if params['experiment']['tracking_dir'] else 'wandb')
            dataset_interface = self._get_dataset_interface()
            self.dataset = dataset_interface(self.dataset_params)
            self.seg_trainer.connect_dataset_interface(self.dataset,
                                                       data_loader_num_workers=params['dataset']['num_workers'])
            self.seg_trainer.init_model(params, False, None)
            self.seg_trainer.init_loggers({"in_params": params}, self.train_params)
            logger.info(f"Input params: \n\n {dict_to_yaml_string(params)}")
        except Exception as e:
            if self.seg_trainer is not None:
                self.seg_trainer.sg_logger.close(True)
            raise e

    def resume(self, wandb_run, updated_config, phases):
        try:
            try:
                self.params = values_to_number(wandb_run.config['in_params'])
            except KeyError:
                raise RuntimeError("No params recorded for run, just delete it!")
            self.params = nested_dict_update(self.params, updated_config)
            self.phases = phases
            wandb_run.config['in_params'] = self.params
            wandb_run.update()
            self.train_params, self.test_params, self.dataset_params, self.early_stop = parse_params(self.params)

            self.seg_trainer = SegmentationTrainer(experiment_name=self.params['experiment']['group'],
                                              ckpt_root_dir=self.params['experiment']['tracking_dir']
                                              if self.params['experiment']['tracking_dir'] else 'wandb')
            dataset_interface = self._get_dataset_interface()
            self.dataset = dataset_interface(self.dataset_params)
            self.seg_trainer.connect_dataset_interface(self.dataset, data_loader_num_workers=self.params['dataset']['num_workers'])
            track_dir = wandb_run.config.get('in_params').get('experiment').get('tracking_dir') or 'wandb'
            checkpoint_path_group = os.path.join(track_dir, wandb_run.group, 'wandb')
            run_folder = list(filter(lambda x: str(wandb_run.id) in x, os.listdir(checkpoint_path_group)))
            checkpoint_path = None
            if 'epoch' in wandb_run.summary:
                ckpt = 'ckpt_latest.pth' if 'train' in phases else 'ckpt_best.pth'
                checkpoint_path = os.path.join(checkpoint_path_group, run_folder[0], 'files', ckpt)
            self.seg_trainer.init_model(self.params, True, checkpoint_path)
            self.seg_trainer.init_loggers({"in_params": self.params}, self.train_params, run_id=wandb_run.id)
        except Exception as e:
            if self.seg_trainer is not None:
                self.seg_trainer.sg_logger.close(really=True)
            raise e

    def _get_dataset_interface(self):
        try:
            dataset_interface = getattr(ezdl.data, self.params['dataset_interface'])
        except AttributeError:
            raise AttributeError("No interface found!")
        return dataset_interface

    def launch(self):
        try:
            if 'train' in self.phases:
                train(self.seg_trainer, self.train_params, self.dataset, self.early_stop)

            if 'test' in self.phases:
                test_metrics = self.seg_trainer.test(**self.test_params)

            if 'inference' in self.phases:
                inference(self.seg_trainer, self.run_params, self.dataset)
        finally:
            if self.seg_trainer is not None:
                self.seg_trainer.sg_logger.close(True)


def train(seg_trainer, train_params, dataset, early_stop):
    # Callbacks
    cbcks = [
        WandbCallback(Phase.TRAIN_EPOCH_END, freq=1),
        WandbCallback(Phase.VALIDATION_EPOCH_END, freq=1),
        SegmentationVisualizationCallback(phase=Phase.VALIDATION_BATCH_END,
                                          freq=1,
                                          batch_idxs=[0, len(seg_trainer.train_loader) - 1],
                                          last_img_idx_in_batch=4,
                                          num_classes=dataset.trainset.CLASS_LABELS,
                                          undo_preprocessing=dataset.undo_preprocess),
        *early_stop
    ]
    train_params["phase_callbacks"] = cbcks

    seg_trainer.train(train_params)
    gc.collect()


def inference(seg_trainer, run_params, dataset):
    run_loader = dataset.get_run_loader(folders=run_params['run_folders'], batch_size=run_params['batch_size'])
    cbcks = [
        # SaveSegmentationPredictionsCallback(phase=Phase.POST_TRAINING,
        #                                     path=
        #                                     run_params['prediction_folder']
        #                                     if run_params['prediction_folder'] != 'mlflow'
        #                                     else mlclient.run.info.artifact_uri + '/predictions',
        #                                     num_classes=len(seg_trainer.test_loader.dataset.classes),
        #                                     )
    ]
    run_loader.dataset.return_name = True
    seg_trainer.run(run_loader, callbacks=cbcks)
    # seg_trainer.valid_loader.dataset.return_name = True
    # seg_trainer.run(seg_trainer.valid_loader, callbacks=cbcks)



import numpy as np
import torch
from super_gradients.training.utils.callbacks import Phase
from super_gradients.training.utils.early_stopping import EarlyStop

from ezdl.learning.wandb_logger import WandBSGLogger
from ezdl.loss import LOSSES as LOSSES_DICT
from ezdl.metrics import metrics_factory


def parse_params(params: dict) -> (dict, dict, dict, list):
    # Set Random seeds
    torch.manual_seed(params['train_params']['seed'])
    np.random.seed(params['train_params']['seed'])

    # Instantiate loss
    input_train_params = params['train_params']
    loss_params = params['train_params']['loss']
    loss = LOSSES_DICT[loss_params['name']](**loss_params['params'])

    # metrics
    train_metrics = metrics_factory(params['train_metrics'])
    test_metrics = metrics_factory(params['test_metrics'])

    # dataset params
    dataset_params = params['dataset']

    train_params = {
        **input_train_params,
        "train_metrics_list": list(train_metrics.values()),
        "valid_metrics_list": list(test_metrics.values()),
        "loss": loss,
        "loss_logging_items_names": ["loss"],
        "sg_logger": WandBSGLogger,
        'sg_logger_params': {
            'entity': params['experiment']['entity'],
            'tags': params['tags'],
            'project_name': params['experiment']['name'],
        }
    }

    test_params = {
        "test_metrics": test_metrics,
    }

    # early stopping
    early_stop = [EarlyStop(Phase.VALIDATION_EPOCH_END, **params['early_stopping']['params'])] \
        if params['early_stopping']['enabled'] else []

    return train_params, test_params, dataset_params, early_stop



import torch
from torch import nn, Tensor
from torch.nn import functional as F
from ezdl.models.layers import ConvModule


class PPM(nn.Module):
    """Pyramid Pooling Module in PSPNet
    """
    def __init__(self, c1, c2=128, scales=(1, 2, 3, 6)):
        super().__init__()
        self.stages = nn.ModuleList([
            nn.Sequential(
                nn.AdaptiveAvgPool2d(scale),
                ConvModule(c1, c2, 1)
            )
        for scale in scales])

        self.bottleneck = ConvModule(c1 + c2 * len(scales), c2, 3, 1, 1)

    def forward(self, x: Tensor) -> Tensor:
        outs = []
        for stage in self.stages:
            outs.append(F.interpolate(stage(x), size=x.shape[-2:], mode='bilinear', align_corners=True))

        outs = [x] + outs[::-1]
        out = self.bottleneck(torch.cat(outs, dim=1))
        return out


if __name__ == '__main__':
    model = PPM(512, 128)
    x = torch.randn(2, 512, 7, 7)   
    y = model(x)
    print(y.shape)  # [2, 128, 7, 7]


import multiprocessing
import os

import PIL.Image
import torch

from tqdm import tqdm
from torchvision.transforms import transforms
from torchvision.transforms import functional as F, InterpolationMode

from ezdl.data.sequoia import WeedMapDataset

WORKERS = multiprocessing.cpu_count()
WEED_CLASS = 2
ROTATIONS_REDEDGE = 0
ROTATIONS_SEQUOIA = 4
CROP_SIZE = [256, 256]
CIR = [3, 1, 0]


def get_crops(img, size):
    if isinstance(img, PIL.Image.Image):
        h, w = img.size
    elif isinstance(img, torch.Tensor):
        w, h = img.size()[-2:]
    else:
        raise TypeError("img should be PIL.Image or torch.Tensor, not {}".format(type(img)))
    yield 0, F.crop(img, 0, 0, *size)
    yield 1, F.crop(img, 0, h - size[1], *size)
    yield 2, F.crop(img, w - size[0], 0, *size)
    yield 3, F.crop(img, w - size[0], h - size[1], *size)


def generate_train(root, target_root, train_folders, channels):
    index = WeedMapDataset.build_index(
        root,
        macro_folders=train_folders,
        channels=channels,
    )

    sq = WeedMapDataset(root,
                        transform=lambda x: x,
                        target_transform=lambda x: x,
                        index=index,
                        channels=channels,
                        return_path=True
                        )

    init_degrees = [-180, 180]

    totensor = transforms.ToTensor()
    # loop through images
    images_with_weed = 0
    for input, target, additional in tqdm(sq):
        for (k, input_crop), (_, target_crop) in zip(get_crops(input, CROP_SIZE), get_crops(target, CROP_SIZE)):
            f, name = additional['input_name'].split('_')
            name = name.split('.')[0]

            if WEED_CLASS in totensor(target_crop).unique():
                for r in range(ROTATIONS):
                    degrees = torch.FloatTensor(1).uniform_(*init_degrees).item()
                    r_input = F.rotate(input_crop, degrees, InterpolationMode.BILINEAR)
                    r_target = F.rotate(target_crop, degrees, InterpolationMode.NEAREST)
                    for i, c in enumerate(r_input):
                        F.to_pil_image(c).save(os.path.join(target_root, f, "tile", channels[i], f"{name}_{k}_{r}.png"))
                    r_target.save(os.path.join(target_root, f, "groundtruth", f"{f}_{name}_{k}_{r}_GroundTruth_iMap.png"))
                    F.to_pil_image(torch.stack([r_input[CIR[0]], r_input[CIR[1]], r_input[CIR[2]]]))\
                        .save(os.path.join(target_root, f, "tile", 'CIR', f"{name}_{k}_{r}.png"))
                images_with_weed += 1

            for i, c in enumerate(input_crop):
                F.to_pil_image(c).save(os.path.join(target_root, f, "tile", channels[i], f"{name}_{k}.png"))
            target_crop.save(os.path.join(target_root, f, "groundtruth", f"{f}_{name}_{k}_GroundTruth_iMap.png"))
            F.to_pil_image(torch.stack([input_crop[CIR[0]], input_crop[CIR[1]], input_crop[CIR[2]]])) \
                .save(os.path.join(target_root, f, "tile", 'CIR', f"{name}_{k}.png"))

    print(f"{images_with_weed} images with weed on {len(sq)} images")  # 363 images with weed


def generate_test(root, target_root, test_folders, channels):
    index = WeedMapDataset.build_index(
        root,
        macro_folders=test_folders,
        channels=channels,
    )

    sq = WeedMapDataset(root,
                        transform=lambda x: x,
                        target_transform=lambda x: x,
                        index=index,
                        channels=channels,
                        return_path=True
                        )
    # loop through images
    for input, target, additional in tqdm(sq):
        for (k, input_crop), (_, target_crop) in zip(get_crops(input, CROP_SIZE), get_crops(target, CROP_SIZE)):
            f, name = additional['input_name'].split('_')
            name = name.split('.')[0]

            for i, c in enumerate(input_crop):
                F.to_pil_image(c).save(os.path.join(target_root, f, "tile", channels[i], f"{name}_{k}.png"))
            target_crop.save(os.path.join(target_root, f, "groundtruth", f"{f}_{name}_{k}_GroundTruth_iMap.png"))
            F.to_pil_image(torch.stack([input_crop[CIR[0]], input_crop[CIR[1]], input_crop[CIR[2]]])) \
                .save(os.path.join(target_root, f, "tile", 'CIR', f"{name}_{k}.png"))


def calculate():
    """
    Calculate the mean and the standard deviation of a dataset
    """
    torch.manual_seed(42)
    global ROTATIONS

    # Sequoia 277 images with weed on 337 (x4) images
    # ROTATIONS = ROTATIONS_SEQUOIA
    # root = "./dataset/processed/Sequoia"
    # train_folders = ['006', '007']
    # test_folders = ['005']
    # target_root = f"./dataset/{ROTATIONS}_rotations_processed_{'-'.join(test_folders)}_test/Sequoia"
    # channels = ['R', 'G', 'NDVI', 'NIR', 'RE']

    # RedEdge 1285 images with weed on 455 (x4) images
    ROTATIONS = ROTATIONS_REDEDGE
    root = "./dataset/processed/RedEdge"
    train_folders = ['000', '001', '002', '004']
    test_folders = ['003']
    target_root = f"./dataset/{ROTATIONS}_rotations_processed_{'-'.join(test_folders)}_test/RedEdge"
    channels = ['R', 'G', 'B', 'NDVI', 'NIR', 'RE']

    os.makedirs(target_root, exist_ok=True)
    for f in train_folders + test_folders:
        os.makedirs(os.path.join(target_root, f), exist_ok=True)
        for c in channels:
            os.makedirs(os.path.join(target_root, f, "tile", c), exist_ok=True)
        os.makedirs(os.path.join(target_root, f, "groundtruth"), exist_ok=True)
        os.makedirs(os.path.join(target_root, f, "tile", "CIR"), exist_ok=True)

    generate_train(root, target_root, train_folders, channels)
    generate_test(root, target_root, test_folders, channels)


if __name__ == '__main__':
    calculate()



import torch
from torch import nn, Tensor
from ezdl.models.layers import DropPath


class MLP(nn.Module):
    def __init__(self, dim, hidden_dim, out_dim=None) -> None:
        super().__init__()
        out_dim = out_dim or dim
        self.fc1 = nn.Linear(dim, hidden_dim)
        self.act = nn.GELU()
        self.fc2 = nn.Linear(hidden_dim, out_dim)

    def forward(self, x: Tensor) -> Tensor:
        return self.fc2(self.act(self.fc1(x)))


class CMLP(nn.Module):
    def __init__(self, dim, hidden_dim, out_dim=None) -> None:
        super().__init__()
        out_dim = out_dim or dim
        self.fc1 = nn.Conv2d(dim, hidden_dim, 1)
        self.act = nn.GELU()
        self.fc2 = nn.Conv2d(hidden_dim, out_dim, 1)

    def forward(self, x: Tensor) -> Tensor:
        return self.fc2(self.act(self.fc1(x)))


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.scale = (dim // num_heads) ** -0.5
        self.qkv = nn.Linear(dim, dim*3)
        self.proj = nn.Linear(dim, dim)

    def forward(self, x: Tensor) -> Tensor:
        B, N, C = x.shape
        q, k, v = self.qkv(x).reshape(B, N, 3, self.num_heads, C//self.num_heads).permute(2, 0, 3, 1, 4)
        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        
        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        return x


class CBlock(nn.Module):
    def __init__(self, dim, dpr=0.):
        super().__init__()
        self.pos_embed = nn.Conv2d(dim, dim, 3, 1, 1, groups=dim)
        self.norm1 = nn.BatchNorm2d(dim)
        self.conv1 = nn.Conv2d(dim, dim, 1)
        self.conv2 = nn.Conv2d(dim, dim, 1)
        self.attn = nn.Conv2d(dim, dim, 5, 1, 2, groups=dim)
        self.drop_path = DropPath(dpr) if dpr > 0. else nn.Identity()
        self.norm2 = nn.BatchNorm2d(dim)
        self.mlp = CMLP(dim, int(dim*4))

    def forward(self, x: Tensor) -> Tensor:
        x = x + self.pos_embed(x)
        x = x + self.drop_path(self.conv2(self.attn(self.conv1(self.norm1(x))))) 
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class SABlock(nn.Module):
    def __init__(self, dim, num_heads, dpr=0.) -> None:
        super().__init__()
        self.pos_embed = nn.Conv2d(dim, dim, 3, 1, 1, groups=dim)
        self.norm1 = nn.LayerNorm(dim)
        self.attn = Attention(dim, num_heads)
        self.drop_path = DropPath(dpr) if dpr > 0. else nn.Identity()
        self.norm2 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, int(dim*4))

    def forward(self, x: Tensor) -> Tensor:
        x = x + self.pos_embed(x)
        B, N, H, W = x.shape
        x = x.flatten(2).transpose(1, 2)
        x = x + self.drop_path(self.attn(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        x = x.transpose(1, 2).reshape(B, N, H, W)
        return x


class PatchEmbed(nn.Module):
    def __init__(self, patch_size=16, in_ch=3, embed_dim=768) -> None:
        super().__init__()
        self.norm = nn.LayerNorm(embed_dim)
        self.proj = nn.Conv2d(in_ch, embed_dim, patch_size, patch_size)

    def forward(self, x: Tensor) -> Tensor:
        x = self.proj(x)
        B, C, H, W = x.shape
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        return x


uniformer_settings = {
    'S': [3, 4, 8, 3],       # [depth]
    'B': [5, 8, 20, 7]
}


class UniFormer(nn.Module):     
    def __init__(self, model_name: str = 'S') -> None:
        super().__init__()
        assert model_name in uniformer_settings.keys(), f"UniFormer model name should be in {list(uniformer_settings.keys())}"
        depth = uniformer_settings[model_name]

        head_dim = 64
        drop_path_rate = 0.
        embed_dims = [64, 128, 320, 512]
    
        for i in range(4):
            self.add_module(f"patch_embed{i+1}", PatchEmbed(4 if i == 0 else 2, 3 if i == 0 else embed_dims[i-1], embed_dims[i]))
            self.add_module(f"norm{i+1}", nn.LayerNorm(embed_dims[i]))

        self.pos_drop = nn.Dropout(0.)
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depth))]
        num_heads = [dim // head_dim for dim in embed_dims]

        self.blocks1 = nn.ModuleList([
            CBlock(embed_dims[0], dpr[i])
        for i in range(depth[0])])
        
        self.blocks2 = nn.ModuleList([
            CBlock(embed_dims[1], dpr[i+depth[0]])
        for i in range(depth[1])])

        self.blocks3 = nn.ModuleList([
            SABlock(embed_dims[2], num_heads[2], dpr[i+depth[0]+depth[1]])
        for i in range(depth[2])])

        self.blocks4 = nn.ModuleList([
            SABlock(embed_dims[3], num_heads[3], dpr[i+depth[0]+depth[1]+depth[2]])
        for i in range(depth[3])])

        
    def forward(self, x: torch.Tensor):
        outs = []

        x = self.patch_embed1(x)
        x = self.pos_drop(x)
        for blk in self.blocks1:
            x = blk(x)
        x_out = self.norm1(x.permute(0, 2, 3, 1))
        outs.append(x_out.permute(0, 3, 1, 2))

        x = self.patch_embed2(x)
        for blk in self.blocks2:
            x = blk(x)
        x_out = self.norm2(x.permute(0, 2, 3, 1))
        outs.append(x_out.permute(0, 3, 1, 2))

        x = self.patch_embed3(x)
        for blk in self.blocks3:
            x = blk(x)
        x_out = self.norm3(x.permute(0, 2, 3, 1))
        outs.append(x_out.permute(0, 3, 1, 2))

        x = self.patch_embed4(x)
        for blk in self.blocks4:
            x = blk(x)
        x_out = self.norm4(x.permute(0, 2, 3, 1))
        outs.append(x_out.permute(0, 3, 1, 2))

        return outs

if __name__ == '__main__':
    model = UniFormer('S')
    model.load_state_dict(torch.load('C:\\Users\\sithu\\Documents\\weights\\backbones\\uniformer\\uniformer_small_in1k.pth', map_location='cpu')['model'], strict=False)
    x = torch.randn(1, 3, 224, 224)
    feats = model(x)
    for y in feats:
        print(y.shape)



import torch
import torch.nn as nn


class Random(nn.Module):
    def __init__(self, arch_params):
        super(Random, self).__init__()
        in_chn, out_chn = arch_params['input_channels'], arch_params['output_channels']

        self.in_chn = in_chn
        self.out_chn = out_chn

    def forward(self, x):
        return torch.rand(x.shape[0], self.out_chn, *(x.shape[2:]), device=x.device)
